import json
from datetime import datetime, timedelta

import pytest

from cg.apps.lims import LimsAPI
from cg.meta.report.api import ReportAPI
from cg.store import Store


@pytest.fixture
def lims_family():
    return json.load(open("tests/fixtures/report/lims_family.json"))


@pytest.fixture
def lims_samples(lims_family):
    return lims_family["samples"]


@pytest.fixture
def report_samples(lims_family):
    for sample in lims_family["samples"]:
        sample["internal_id"] = sample["id"]

    return lims_family["samples"]


class MockLims(LimsAPI):
    def __init__(self, samples):
        self._samples = samples

    def get_prep_method(self, lims_id: str) -> str:
        return (
            "CG002 - End repair Size selection A-tailing and Adapter ligation (TruSeq PCR-free "
            ""
            "DNA)"
        )

    def get_sequencing_method(self, lims_id: str) -> str:
        return "CG002 - Cluster Generation (HiSeq X)"

    def get_delivery_method(self, lims_id: str) -> str:
        return "CG002 - Delivery"

    def sample(self, lims_id: str):
        """Fetch information about a sample."""

        for sample in self._samples:
            if sample.get("id") == lims_id:
                return sample

        return None


class MockFile:
    def __init__(self, path):
        self.path = path


class MockChanjo:
    _sample_coverage_returns_none = False

    def sample_coverage(self, sample_id: str, panel_genes: list) -> dict:
        """Calculate coverage for OMIM panel."""

        if self._sample_coverage_returns_none:
            return None

        if sample_id == "ADM1":
            data = {"mean_coverage": 38.342, "mean_completeness": 99.1}
        elif sample_id == "ADM2":
            data = {"mean_coverage": 37.342, "mean_completeness": 97.1}
        else:
            data = {"mean_coverage": 39.342, "mean_completeness": 98.1}

        return data

    def sample(self, sample_id: str) -> dict:
        """Fetch sample from the database."""
        return {"id": sample_id}


class MockPath:
    _path = None

    def __call__(self, path):
        self._path = path
        return self

    def joinpath(self, another_path):
        return self(self._path + another_path)

    def open(self):
        pass


class MockAnalysis:
    def panel(self, family_obj) -> [str]:
        """Create the aggregated panel file."""
        return [""]

    def get_latest_metadata(self, family_id):
        # Returns: dict: parsed data
        # Define output dict
        out_data = {
            "analysis_sex": {"ADM1": "female", "ADM2": "female", "ADM3": "female"},
            "family": "yellowhog",
            "duplicates": {"ADM1": 13.525, "ADM2": 12.525, "ADM3": 14.525},
            "genome_build": "hg19",
            "mapped_reads": {"ADM1": 98.8, "ADM2": 99.8, "ADM3": 97.8},
            "mip_version": "v4.0.20",
            "sample_ids": ["2018-20203", "2018-20204"],
        }

        return out_data


class MockLogger:
    last_warning = None
    warnings = []

    def warning(self, text, *interpolations):

        self.warnings.append(text % interpolations)
        self.last_warning = text % interpolations

    def get_last_warning(self) -> str:
        return self.last_warning

    def get_warnings(self) -> list:
        return self.warnings


class MockYamlLoader:
    def __init__(self):
        self.open_file = None

    def safe_load(self, open_file: str):
        self.open_file = open_file
        return "safely_loaded_file"

    def get_open_file(self):
        return self.open_file


class MockDB(Store):
    family_samples_returns_no_reads = False
    samples_returns_no_capture_kit = False
    _application_accreditation = None

    def __init__(self, store):
        self.store = store

    def family_samples(self, family_id: str):

        family_samples = self.store.family_samples(family_id)

        if self.family_samples_returns_no_reads:
            for family_sample in family_samples:
                family_sample.sample.reads = None

        if self.samples_returns_no_capture_kit:
            for family_sample in family_samples:
                family_sample.sample.capture_kit = None

        # add some date to calculate processing time on
        yesterday = datetime.now() - timedelta(days=1)
        for family_sample in family_samples:
            family_sample.sample.received_at = yesterday
            family_sample.sample.prepared_at = yesterday
            family_sample.sample.sequenced_at = yesterday
            family_sample.sample.delivered_at = datetime.now()

        return family_samples

    def application(self, tag: str):
        """Fetch an application from the store."""
        application = self.store.application(tag=tag)

        if self._application_accreditation is not None:
            application.is_accredited = self._application_accreditation

        return application


class MockScout:
    def get_genes(self, panel_id: str, version: str = None) -> list:
        return []


@pytest.fixture(scope="function")
def report_store(analysis_store, helpers):
    family = analysis_store.families()[0]
    helpers.add_analysis(analysis_store, family)
    helpers.add_analysis(analysis_store, family)
    return analysis_store


@pytest.fixture(scope="function")
def report_api(report_store, lims_samples):
    db = MockDB(report_store)
    lims = MockLims(lims_samples)
    chanjo = MockChanjo()
    analysis = MockAnalysis()
    scout = MockScout()
    logger = MockLogger()
    yaml_loader = MockYamlLoader()
    path_tool = MockPath()
    _report_api = ReportAPI(
        lims_api=lims,
        store=db,
        chanjo_api=chanjo,
        analysis_api=analysis,
        scout_api=scout,
        logger=logger,
        yaml_loader=yaml_loader,
        path_tool=path_tool,
    )
    return _report_api



"""Hold sbatch templates in global variables"""

SBATCH_HEADER_TEMPLATE = """#! /bin/bash
#SBATCH --job-name={job_name}
#SBATCH --account={account}
#SBATCH --ntasks={ntasks}
#SBATCH --mem={mem}G
#SBATCH --error={log_dir}/{job_name}.stderr
#SBATCH --output={log_dir}/{job_name}.stdout
#SBATCH --mail-type=FAIL
#SBATCH --mail-user={mail_user}
#SBATCH --time={time}:00:00
#SBATCH --qos=low

set -e

echo "Running on: $(hostname)"

source activate {conda_env}
"""

######################################################################

SBATCH_FASTQ_TO_SPRING = """
error() {{
    if [[ -e {spring_path} ]]
    then
        rm {spring_path}
    fi

    if [[ -e {pending_path} ]]
    then
        rm {pending_path}
    fi

    exit 1
}}

trap error ERR

mkdir -p {tmp_dir}
crunchy -t 12 --tmp-dir {tmp_dir} compress fastq -f {fastq_first} -s {fastq_second} \
-o {spring_path} --check-integrity --metadata-file
rm {pending_path}
rm -r {tmp_dir}"""

######################################################################

SBATCH_SPRING_TO_FASTQ = """
error() {{
    if [[ -e {fastq_first} ]]
    then
        rm {fastq_first}
    fi

    if [[ -e {fastq_second} ]]
    then
        rm {fastq_second}
    fi

    if [[ -e {pending_path} ]]
    then
        rm {pending_path}
    fi

    exit 1
}}

trap error ERR

mkdir -p {tmp_dir}
crunchy -t 12 --tmp-dir {tmp_dir} decompress spring {spring_path} -f {fastq_first} -s \
{fastq_second} --first-checksum {checksum_first} --second-checksum {checksum_second}
rm {pending_path}
rm -r {tmp_dir}"""



"""Import functionality"""

import logging
import sys
from datetime import datetime

import xlrd
from cg.exc import CgError
from cg.store import models, Store


def import_application_versions(store, excel_path, sign, dry_run, skip_missing):
    """
    Imports all application versions from the specified excel file
    Args:
        :param store:               Status database store
        :param excel_path:          Path to excel file with headers: 'App tag', 'Valid from',
                                    'Standard', 'Priority', 'Express', 'Research'
        :param sign:                Signature of user running the script
        :param dry_run:             Test run, no changes to the database
        :param skip_missing:        Continue despite missing applications
    """

    workbook = XlFileHelper.get_workbook_from_xl(excel_path)
    raw_versions = XlFileHelper.get_raw_dicts_from_xl(excel_path)

    for raw_version in raw_versions:
        tag = _get_tag_from_raw_version(raw_version)
        application_obj = store.application(tag)

        if not application_obj:
            logging.error("Failed to find application! Please manually " "add application: %s", tag)

            if skip_missing:
                continue

            logging.error("Rolling back transaction.")
            store.rollback()
            sys.exit()

        app_tag = application_obj.tag
        latest_version = store.latest_version(tag)

        if latest_version and versions_are_same(latest_version, raw_version, workbook.datemode):
            logging.info("skipping redundant application version for app tag %s", app_tag)
            continue

        logging.info("adding new application version to transaction for app tag %s", app_tag)
        new_version = add_version_from_raw(
            application_obj, latest_version, raw_version, sign, store, workbook
        )
        store.add(new_version)

    if not dry_run:
        logging.info(
            "all application versions successfully added to transaction, committing " "transaction"
        )
        store.commit()
    else:
        logging.error("Dry-run, rolling back transaction.")
        store.rollback()


def import_applications(store, excel_path, sign, dry_run, sheet_name=None):
    """
    Imports all applications from the specified excel file
    Args:
        :param sheet_name:          Optional sheet name
        :param store:               status database store
        :param excel_path:          Path to excel file with headers: 'App tag', 'Valid from',
                                    'Standard', 'Priority', 'Express', 'Research'
        :param sign:                Signature of user running the script
        :param dry_run:             Test run, no changes to the database
    """

    raw_applications = XlFileHelper.get_raw_dicts_from_xl(excel_path, sheet_name)

    for raw_application in raw_applications:
        tag = _get_tag_from_raw_application(raw_application)
        application_obj = store.application(tag)

        if application_obj and applications_are_same(application_obj, raw_application):
            logging.info("skipping redundant application %s", tag)
            continue

        logging.info("adding new application to transaction %s", tag)
        new_application = add_application_from_raw(raw_application, sign, store)
        store.add(new_application)

    if not dry_run:
        logging.info("all applications successfully added to transaction, committing transaction")
        store.commit()
    else:
        logging.error("Dry-run, rolling back transaction.")
        store.rollback()


def prices_are_same(first_price, second_price):
    """Checks if the given prices are to be considered equal"""

    if first_price == second_price:
        return True
    if first_price and second_price:
        return int(round(float(first_price))) == int(round(float(second_price)))

    return False


def versions_are_same(version_obj: models.ApplicationVersion, raw_version: dict, datemode: int):
    """Checks if the given versions are to be considered equal"""

    return (
        version_obj.application.tag == _get_tag_from_raw_version(raw_version)
        and version_obj.valid_from
        == datetime(*xlrd.xldate_as_tuple(float(raw_version["Valid from"]), datemode))
        and prices_are_same(version_obj.price_standard, raw_version["Standard"])
        and prices_are_same(version_obj.price_priority, raw_version["Priority"])
        and prices_are_same(version_obj.price_express, raw_version["Express"])
        and prices_are_same(version_obj.price_research, raw_version["Research"])
    )


def applications_are_same(application_obj: models.Application, raw_application: dict):
    """Checks if the given applications are to be considered equal"""

    return application_obj and application_obj.tag == _get_tag_from_raw_application(raw_application)


def add_version_from_raw(
    application_obj, latest_version, raw_version, sign, store: Store, workbook
):
    """Adds an application version from a raw application version record"""
    new_version = store.add_version(
        application=application_obj,
        version=latest_version.version + 1 if latest_version else 1,
        valid_from=datetime(
            *xlrd.xldate_as_tuple(float(raw_version["Valid from"]), workbook.datemode)
        ),
        prices={
            "standard": raw_version["Standard"],
            "priority": raw_version["Priority"],
            "express": raw_version["Express"],
            "research": raw_version["Research"],
        },
        comment="Added by %s" % sign,
    )
    return new_version


def add_application_from_raw(raw_application, sign, store: Store):
    """Adds an application from a raw application record"""
    new_application = store.add_application(
        tag=_get_tag_from_raw_application(raw_application),
        category=raw_application["prep_category"],
        description=raw_application["description"],
        is_accredited=raw_application["is_accredited"] == 1.0,
        turnaround_time=raw_application["turnaround_time"],
        minimum_order=raw_application["minimum_order"],
        sequencing_depth=raw_application["sequencing_depth"],
        target_reads=raw_application["target_reads"],
        sample_amount=raw_application["sample_amount"],
        sample_volume=raw_application["sample_volume"],
        sample_concentration=raw_application["sample_concentration"],
        priority_processing=raw_application["priority_processing"] == 1.0,
        details=raw_application["details"],
        limitations=raw_application["limitations"],
        percent_kth=raw_application["percent_kth"],
        comment=raw_application["comment"] + " Added by %s" % sign,
        is_archived=raw_application["is_archived"] == 1.0,
        is_external=raw_application["is_external"] == 1.0,
        created_at=datetime.now(),
    )
    return new_application


def import_apptags(
    store: Store,
    excel_path: str,
    prep_category: str,
    sign: str,
    sheet_name: str,
    tag_column: int,
    activate: bool,
    inactivate: bool,
):
    """Syncs all applications from the specified excel file"""

    orderform_application_tags = []

    for raw_row in XlFileHelper.get_raw_cells_from_xl(excel_path, sheet_name, tag_column):
        tag = _get_tag_from_column(raw_row, tag_column)
        logging.info("Found: %s in orderform", tag)
        orderform_application_tags.append(tag)

    if not orderform_application_tags:
        message = "No applications found in column %s (zero-based), exiting" % tag_column
        raise CgError(message)

    for orderform_application_tag in orderform_application_tags:
        application_obj = store.application(tag=orderform_application_tag)

        if not application_obj:
            message = "Application %s was not found" % orderform_application_tag
            raise CgError(message)

        if application_obj.prep_category != prep_category:
            message = "%s prep_category, expected: %s was: %s" % (
                orderform_application_tag,
                prep_category,
                application_obj.prep_category,
            )
            raise CgError(message)

        if application_obj.is_archived:
            if activate:
                application_obj.comment = (
                    f"{application_obj.comment}\n{str(datetime.now())[:-10]}"
                    f"Application un-archived by {sign}"
                )
                application_obj.is_archived = False
                logging.info("Un-archiving %s", application_obj)
            else:
                logging.warning(
                    "%s is marked as archived but is used in the orderform, consider "
                    "activating it",
                    application_obj,
                )
        else:
            logging.info("%s is already active, no need to activate it", application_obj)

    all_active_apps_for_category = store.applications(category=prep_category, archived=False)

    for active_application in all_active_apps_for_category:
        if active_application.tag not in orderform_application_tags:
            if inactivate:
                active_application.is_archived = True
                active_application.comment = (
                    f"{active_application.comment}"
                    f"\n{str(datetime.now())[:-10]} "
                    f"Application archived by {sign}"
                )
                logging.info("Archiving %s", active_application)
            else:
                logging.warning(
                    "%s is marked as active but is not used in the orderform, "
                    "consider archiving it",
                    active_application,
                )
        else:
            logging.info(
                "%s was found in orderform tags, no need to archive it", active_application
            )

    if not activate and not inactivate:
        logging.info("no change mode requested, rolling back transaction")
        store.rollback()
    else:
        logging.info("all applications successfully synced, committing transaction")
        store.commit()


class XlFileHelper:
    """Organises methods to get data from file"""

    @staticmethod
    def get_raw_dicts_from_xl(excel_path, sheet_name=None):
        """Get raw data from the xl file"""
        workbook = XlFileHelper.get_workbook_from_xl(excel_path)
        if sheet_name:
            data_sheet = workbook.sheet_by_name(sheet_name)
        else:
            data_sheet = workbook.sheet_by_index(0)
        return XlSheetHelper.get_raw_dicts_from_xlsheet(data_sheet)

    @staticmethod
    def get_raw_cells_from_xl(excel_path, sheet_name, tag_column):
        """Get raw data from the xl file"""
        workbook = XlFileHelper.get_workbook_from_xl(excel_path)
        data_sheet = workbook.sheet_by_name(sheet_name)
        return XlSheetHelper.get_raw_cells_from_sheet(data_sheet, tag_column)

    @staticmethod
    def get_workbook_from_xl(excel_path):
        """Get the workbook from the xl file"""
        return xlrd.open_workbook(excel_path)

    @staticmethod
    def get_datemode_from_xl(excel_path):
        """"Returns the datemode of of the workbook in the specified xl"""
        workbook = xlrd.open_workbook(excel_path)
        return workbook.datemode


class XlSheetHelper:
    """Organises methods to get data from sheet"""

    @staticmethod
    def get_raw_dicts_from_xlsheet(sheet) -> []:
        """Get the relevant rows from a sheet."""
        raw_data = []
        header_row = []
        first_row = True

        for row in sheet.get_rows():
            if row[0].value == "":
                break

            if first_row:
                header_row = [cell.value for cell in row]
                first_row = False
            else:
                values = [str(cell.value) for cell in row]
                version_dict = dict(zip(header_row, values))
                raw_data.append(version_dict)

        return raw_data

    @staticmethod
    def no_more_tags(row, tag_column: int) -> bool:
        """Returns if there are no more app-tags found"""
        return _get_tag_from_column(row, tag_column).value == ""

    @staticmethod
    def get_raw_cells_from_sheet(sheet, tag_column: int) -> []:
        """Get the relevant rows from an price sheet."""
        raw_data = []

        for row in sheet.get_rows():
            if XlSheetHelper.no_more_tags(row, tag_column):
                break

            values = [str(cell.value) for cell in row]
            raw_data.append(values)

        return raw_data


def _get_tag_from_raw_version(raw_data):
    """Gets the application tag from a raw xl application version record"""
    return raw_data["App tag"]


def _get_tag_from_column(raw_data, column: int):
    """Gets the application tag from a raw xl data record"""
    return raw_data[column]


def _get_tag_from_raw_application(raw_application):
    """Gets the application tag from a raw xl application record"""
    return raw_application["tag"]



import logging
from pathlib import Path
from typing import Iterator

import alchy
import sqlalchemy as sqa

from cgstats.db import api, models

LOG = logging.getLogger(__name__)


class StatsAPI(alchy.Manager):

    Project = models.Project
    Sample = models.Sample
    Unaligned = models.Unaligned
    Supportparams = models.Supportparams
    Datasource = models.Datasource
    Demux = models.Demux
    Flowcell = models.Flowcell

    def __init__(self, config: dict):
        alchy_config = dict(SQLALCHEMY_DATABASE_URI=config["cgstats"]["database"])
        super(StatsAPI, self).__init__(config=alchy_config, Model=models.Model)
        self.root_dir = Path(config["cgstats"]["root"])

    def flowcell(self, flowcell_name: str) -> dict:
        """Fetch information about a flowcell."""
        record = self.Flowcell.query.filter_by(flowcellname=flowcell_name).first()
        data = {
            "name": record.flowcellname,
            "sequencer": record.demux[0].datasource.machine,
            "sequencer_type": record.hiseqtype,
            "date": record.time,
            "samples": [],
        }
        for sample_obj in self.flowcell_samples(record):
            raw_samplename = sample_obj.samplename.split("_", 1)[0]
            curated_samplename = raw_samplename.rstrip("AB")
            sample_data = {"name": curated_samplename, "reads": 0, "fastqs": []}
            for fc_data in self.sample_reads(sample_obj):
                if fc_data.type == "hiseqga" and fc_data.q30 >= 80:
                    sample_data["reads"] += fc_data.reads
                elif fc_data.type == "hiseqx" and fc_data.q30 >= 75:
                    sample_data["reads"] += fc_data.reads
                elif fc_data.type == "novaseq" and fc_data.q30 >= 75:
                    sample_data["reads"] += fc_data.reads
                else:
                    LOG.warning(
                        f"q30 too low for {curated_samplename} on {fc_data.name}:"
                        f"{fc_data.q30} < {80 if fc_data.type == 'hiseqga' else 75}%"
                    )
                    continue
                for fastq_path in self.fastqs(fc_data.name, sample_obj):
                    if self.is_lane_pooled(flowcell_obj=record, lane=fc_data.lane):
                        if "Undetermined" in str(fastq_path):
                            continue
                    sample_data["fastqs"].append(str(fastq_path))
            data["samples"].append(sample_data)

        return data

    def flowcell_samples(self, flowcell_obj: models.Flowcell) -> Iterator[models.Sample]:
        """Fetch all the samples from a flowcell."""
        return self.Sample.query.join(models.Sample.unaligned, models.Unaligned.demux).filter(
            models.Demux.flowcell == flowcell_obj
        )

    def is_lane_pooled(self, flowcell_obj: models.Flowcell, lane: str) -> bool:
        """Check whether a lane is pooled or not."""
        query = (
            self.session.query(sqa.func.count(models.Unaligned.sample_id).label("sample_count"))
            .join(models.Unaligned.demux)
            .filter(models.Demux.flowcell == flowcell_obj)
            .filter(models.Unaligned.lane == lane)
        )
        return query.first().sample_count > 1

    def sample_reads(self, sample_obj: models.Sample) -> Iterator:
        """Calculate reads for a sample."""
        query = (
            self.session.query(
                models.Flowcell.flowcellname.label("name"),
                models.Flowcell.hiseqtype.label("type"),
                models.Unaligned.lane,
                sqa.func.sum(models.Unaligned.readcounts).label("reads"),
                sqa.func.min(models.Unaligned.q30_bases_pct).label("q30"),
            )
            .join(models.Flowcell.demux, models.Demux.unaligned)
            .filter(models.Unaligned.sample == sample_obj)
            .group_by(models.Flowcell.flowcellname)
        )
        return query

    def sample(self, sample_name: str) -> models.Sample:
        """Fetch a sample for the database by name."""
        return api.get_sample(sample_name).first()

    def fastqs(self, flowcell: str, sample_obj: models.Sample) -> Iterator[Path]:
        """Fetch FASTQ files for a sample."""
        base_pattern = "*{}/Unaligned*/Project_*/Sample_{}/*.fastq.gz"
        alt_pattern = "*{}/Unaligned*/Project_*/Sample_{}_*/*.fastq.gz"
        for fastq_pattern in (base_pattern, alt_pattern):
            pattern = fastq_pattern.format(flowcell, sample_obj.samplename)
            files = self.root_dir.glob(pattern)
            yield from files



""" Tests for common functions of the compress API """

from cg.meta.compress import files


def test_get_latest_version_no_housekeeper_files(compress_api, caplog):
    """test get_bam_dict method when there is not case in scout"""

    case_id = "compress_case"
    # GIVEN a case id

    # WHEN getting the version_obj
    res = compress_api.get_latest_version(case_id)

    # THEN assert that None was returned since the case was not found in scout
    assert res is None
    # THEN assert that the correct information is being processed
    assert f"No bundle found for {case_id} in housekeeper" in caplog.text



"""
    Conftest file for pytest fixtures that needs to be shared for multiple tests
"""
import copy
import datetime as dt
import logging
import shutil
import os
from pathlib import Path

import pytest
import ruamel.yaml

from cg.apps.hk import HousekeeperAPI
from cg.apps.mip import parse_sampleinfo, parse_qcmetrics
from cg.apps.gt import GenotypeAPI
from cg.constants import Pipeline
from cg.models import CompressionData
from cg.store import Store

from .mocks.crunchy import MockCrunchyAPI
from .mocks.hk_mock import MockHousekeeperAPI
from .mocks.madeline import MockMadelineAPI
from .mocks.scout import MockScoutAPI
from .small_helpers import SmallHelpers
from .store_helpers import StoreHelpers
from .mocks.tb_mock import MockTB
from .mocks.process_mock import ProcessMock

CHANJO_CONFIG = {"chanjo": {"config_path": "chanjo_config", "binary_path": "chanjo"}}
CRUNCHY_CONFIG = {
    "crunchy": {
        "cram_reference": "/path/to/fasta",
        "slurm": {"account": "mock_account", "mail_user": "mock_mail", "conda_env": "mock_env"},
    }
}


LOG = logging.getLogger(__name__)

# Case fixtures


@pytest.fixture(name="case_id")
def fixture_case_id() -> str:
    """Return a case id"""
    return "yellowhog"


@pytest.fixture(name="family_name")
def fixture_family_name() -> str:
    """Return a family name"""
    return "family"


@pytest.fixture(name="customer_id")
def fixture_customer_id() -> str:
    """Return a customer id"""
    return "cust000"


@pytest.fixture(name="ticket_nr")
def fixture_ticket_nr() -> int:
    """Return a ticket nr"""
    return 123456


@pytest.fixture(scope="function", name="analysis_family_single_case")
def fixture_analysis_family_single(case_id: str, family_name: str, ticket_nr: int) -> dict:
    """Build an example family."""
    family = {
        "name": family_name,
        "internal_id": case_id,
        "data_analysis": str(Pipeline.MIP_DNA),
        "application_type": "wgs",
        "panels": ["IEM", "EP"],
        "samples": [
            {
                "name": "proband",
                "sex": "male",
                "internal_id": "ADM1",
                "status": "affected",
                "ticket_number": ticket_nr,
                "reads": 5000000000,
                "capture_kit": "GMSmyeloid",
            }
        ],
    }
    return family


@pytest.yield_fixture(scope="function", name="analysis_family")
def fixture_analysis_family(case_id: str, family_name: str, ticket_nr: int) -> dict:
    """Return a dictionary with information from a analysis family"""
    family = {
        "name": family_name,
        "internal_id": case_id,
        "data_analysis": str(Pipeline.MIP_DNA),
        "application_type": "wgs",
        "panels": ["IEM", "EP"],
        "samples": [
            {
                "name": "child",
                "sex": "male",
                "internal_id": "ADM1",
                "father": "ADM2",
                "mother": "ADM3",
                "status": "affected",
                "ticket_number": ticket_nr,
                "reads": 5000000,
                "capture_kit": "GMSmyeloid",
            },
            {
                "name": "father",
                "sex": "male",
                "internal_id": "ADM2",
                "status": "unaffected",
                "ticket_number": ticket_nr,
                "reads": 6000000,
                "capture_kit": "GMSmyeloid",
            },
            {
                "name": "mother",
                "sex": "female",
                "internal_id": "ADM3",
                "status": "unaffected",
                "ticket_number": ticket_nr,
                "reads": 7000000,
                "capture_kit": "GMSmyeloid",
            },
        ],
    }

    return family


# Config fixtures


@pytest.fixture
def chanjo_config_dict():
    """Chanjo configs"""
    _config = dict()
    _config.update(CHANJO_CONFIG)
    return _config


@pytest.fixture
def crunchy_config_dict():
    """Crunchy configs"""
    _config = dict()
    _config.update(CRUNCHY_CONFIG)
    return _config


@pytest.fixture(name="hk_config_dict")
def fixture_hk_config_dict(root_path):
    """Crunchy configs"""
    _config = {"housekeeper": {"database": "sqlite:///:memory:", "root": str(root_path)}}
    return _config


@pytest.fixture(name="genotype_config")
def fixture_genotype_config() -> dict:
    """
    genotype config fixture
    """
    _config = {
        "genotype": {
            "database": "database",
            "config_path": "config/path",
            "binary_path": "gtdb",
        }
    }
    return _config


# Api fixtures


@pytest.fixture(name="genotype_api")
def fixture_genotype_api(genotype_config: dict) -> GenotypeAPI:
    """
    genotype API fixture
    """
    _genotype_api = GenotypeAPI(genotype_config)
    _genotype_api.set_dry_run(True)
    return _genotype_api


@pytest.yield_fixture(scope="function")
def madeline_api(madeline_output):
    """madeline_api fixture"""
    _api = MockMadelineAPI()
    _api.set_outpath(madeline_output)

    yield _api


# Files fixtures

# Common file fixtures
@pytest.fixture(name="fixtures_dir")
def fixture_fixtures_dir() -> Path:
    """Return the path to the fixtures dir"""
    return Path("tests/fixtures")


@pytest.fixture(name="analysis_dir")
def fixture_analysis_dir(fixtures_dir) -> Path:
    """Return the path to the analysis dir"""
    return fixtures_dir / "analysis"


@pytest.fixture(name="apps_dir")
def fixture_apps_dir(fixtures_dir: Path) -> Path:
    """Return the path to the apps dir"""
    return fixtures_dir / "apps"


@pytest.fixture(name="fastq_dir")
def fixture_fastq_dir(fixtures_dir: Path) -> Path:
    """Return the path to the fastq files dir"""
    return fixtures_dir / "fastq"


@pytest.fixture(scope="function", name="project_dir")
def fixture_project_dir(tmpdir_factory):
    """Path to a temporary directory where intermediate files can be stored"""
    my_tmpdir = Path(tmpdir_factory.mktemp("data"))
    yield my_tmpdir
    shutil.rmtree(str(my_tmpdir))


@pytest.fixture(scope="function")
def tmp_file(project_dir):
    """Get a temp file"""
    return project_dir / "test"


@pytest.fixture(name="orderforms")
def fixture_orderform(fixtures_dir: Path) -> Path:
    """Return the path to the directory with orderforms"""
    _path = fixtures_dir / "orderforms"
    return _path


@pytest.fixture(name="mip_dna_store_files")
def fixture_mip_dna_store_files(apps_dir: Path) -> Path:
    """Return the path to the directory with mip dna store files"""
    _path = apps_dir / "mip" / "dna" / "store"
    return _path


@pytest.fixture(name="mip_analysis_dir")
def fixture_mip_analysis_dir(analysis_dir: Path) -> Path:
    """Return the path to the directory with mip analysis files"""
    _path = analysis_dir / "mip"
    return _path


@pytest.fixture(name="mip_dna_analysis_dir")
def fixture_mip_dna_analysis_dir(mip_analysis_dir: Path) -> Path:
    """Return the path to the directory with mip dna analysis files"""
    _path = mip_analysis_dir / "dna"
    return _path


@pytest.fixture(name="sample1_cram")
def fixture_sample1_cram(mip_dna_analysis_dir: Path) -> Path:
    """Return the path to the cram file for sample 1"""
    _path = mip_dna_analysis_dir / "adm1.cram"
    return _path


@pytest.fixture(name="mip_deliverables_file")
def fixture_mip_deliverables_files(mip_dna_store_files: Path) -> Path:
    """Fixture for general deliverables file in mip"""
    return mip_dna_store_files / "case_id_deliverables.yaml"


@pytest.fixture(name="vcf_file")
def fixture_vcf_file(mip_dna_store_files: Path) -> Path:
    """Return the path to to a vcf file"""
    _path = mip_dna_store_files / "yellowhog_clinical_selected.vcf"
    return _path


# Orderform fixtures


@pytest.fixture
def microbial_orderform(orderforms: Path) -> str:
    """Orderform fixture for microbial samples"""
    _file = orderforms / "1603.9.microbial.xlsx"
    return str(_file)


@pytest.fixture
def rml_orderform():
    """Orderform fixture for RML samples"""
    return "tests/fixtures/orderforms/1604.9.rml.xlsx"


@pytest.fixture(name="madeline_output")
def fixture_madeline_output(apps_dir: Path) -> str:
    """File with madeline output"""
    _file = apps_dir / "madeline/madeline.xml"
    return str(_file)


# Compression fixtures


@pytest.fixture(scope="function", name="run_name")
def fixture_run_name() -> str:
    """Return the name of a fastq run"""
    return "fastq_run"


@pytest.fixture(scope="function", name="original_fastq_data")
def fixture_original_fastq_data(fastq_dir: Path, run_name) -> CompressionData:
    """Return a compression object with a path to the original fastq files"""

    return CompressionData(fastq_dir / run_name)


@pytest.fixture(scope="function", name="fastq_stub")
def fixture_fastq_stub(project_dir: Path, run_name: str) -> Path:
    """Creates a path to the base format of a fastq run"""
    return project_dir / run_name


@pytest.fixture(scope="function", name="compression_object")
def fixture_compression_object(
    fastq_stub: Path, original_fastq_data: CompressionData
) -> CompressionData:
    """Creates compression data object with information about files used in fastq compression"""
    working_files = CompressionData(fastq_stub)
    shutil.copy(str(original_fastq_data.fastq_first), str(working_files.fastq_first))
    shutil.copy(str(original_fastq_data.fastq_second), str(working_files.fastq_second))
    return working_files


# Unknown file fixtures


@pytest.fixture(name="case_qc_metrics")
def fixture_case_qc_metrics(apps_dir: Path) -> Path:
    """Return the path to a qc metrics file with case data"""
    return Path("tests/fixtures/apps/mip/case_qc_metrics.yaml")


@pytest.fixture(name="bcf_file")
def fixture_bcf_file(apps_dir: Path) -> Path:
    """Return the path to a bcf file"""
    return apps_dir / "gt" / "yellowhog.bcf"


@pytest.fixture(scope="session", name="files")
def fixture_files():
    """Trailblazer api for mip files"""
    return {
        "config": "tests/fixtures/apps/mip/dna/store/case_config.yaml",
        "sampleinfo": "tests/fixtures/apps/mip/dna/store/case_qc_sample_info.yaml",
        "qcmetrics": "tests/fixtures/apps/mip/case_qc_metrics.yaml",
        "rna_config": "tests/fixtures/apps/mip/rna/case_config.yaml",
        "rna_sampleinfo": "tests/fixtures/apps/mip/rna/case_qc_sampleinfo.yaml",
        "rna_config_store": "tests/fixtures/apps/mip/rna/store/case_config.yaml",
        "rna_sampleinfo_store": "tests/fixtures/apps/mip/rna/store/case_qc_sample_info.yaml",
        "mip_rna_deliverables": "test/fixtures/apps/mip/rna/store/case_deliverables.yaml",
        "dna_config_store": "tests/fixtures/apps/mip/dna/store/case_config.yaml",
        "dna_sampleinfo_store": "tests/fixtures/apps/mip/dna/store/case_qc_sample_info.yaml",
        "mip_dna_deliverables": "test/fixtures/apps/mip/dna/store/case_deliverables.yaml",
    }


@pytest.fixture(scope="function", name="bed_file")
def fixture_bed_file(analysis_dir) -> str:
    """Get the path to a bed file file"""
    return str(analysis_dir / "sample_coverage.bed")


@pytest.fixture(scope="session", name="files_raw")
def fixture_files_raw(files):
    """Get some raw files"""
    return {
        "config": ruamel.yaml.safe_load(open(files["config"])),
        "sampleinfo": ruamel.yaml.safe_load(open(files["sampleinfo"])),
        "qcmetrics": ruamel.yaml.safe_load(open(files["qcmetrics"])),
        "rna_config": ruamel.yaml.safe_load(open(files["rna_config"])),
        "rna_sampleinfo": ruamel.yaml.safe_load(open(files["rna_sampleinfo"])),
        "rna_config_store": ruamel.yaml.safe_load(open(files["rna_config_store"])),
        "rna_sampleinfo_store": ruamel.yaml.safe_load(open(files["rna_sampleinfo_store"])),
        "dna_config_store": ruamel.yaml.safe_load(open(files["dna_config_store"])),
        "dna_sampleinfo_store": ruamel.yaml.safe_load(open(files["dna_sampleinfo_store"])),
    }


@pytest.fixture(scope="session")
def files_data(files_raw):
    """Get some data files"""
    return {
        "config": parse_sampleinfo.parse_config(files_raw["config"]),
        "sampleinfo": parse_sampleinfo.parse_sampleinfo(files_raw["sampleinfo"]),
        "qcmetrics": parse_qcmetrics.parse_qcmetrics(files_raw["qcmetrics"]),
        "rna_config": parse_sampleinfo.parse_config(files_raw["rna_config"]),
        "rna_sampleinfo": parse_sampleinfo.parse_sampleinfo_rna(files_raw["rna_sampleinfo"]),
        "rna_config_store": parse_sampleinfo.parse_config(files_raw["rna_config_store"]),
        "rna_sampleinfo_store": parse_sampleinfo.parse_sampleinfo(
            files_raw["rna_sampleinfo_store"]
        ),
        "dna_config_store": parse_sampleinfo.parse_config(files_raw["dna_config_store"]),
        "dna_sampleinfo_store": parse_sampleinfo.parse_sampleinfo(
            files_raw["dna_sampleinfo_store"]
        ),
    }


# Helper fixtures


@pytest.fixture(name="helpers")
def fixture_helpers():
    """Return a class with helper functions for the stores"""
    return StoreHelpers()


@pytest.fixture(name="small_helpers")
def fixture_small_helpers():
    """Return a class with small helper functions"""
    return SmallHelpers()


# HK fixtures


@pytest.fixture(name="root_path")
def fixture_root_path(project_dir: Path) -> Path:
    """Return the path to a hk bundles dir"""
    _root_path = project_dir / "bundles"
    _root_path.mkdir(parents=True, exist_ok=True)
    return _root_path


@pytest.fixture(scope="function", name="timestamp")
def fixture_timestamp() -> dt.datetime:
    """Return a time stamp in date time format"""
    return dt.datetime(2020, 5, 1)


@pytest.fixture(scope="function", name="later_timestamp")
def fixture_later_timestamp() -> dt.datetime:
    """Return a time stamp in date time format"""
    return dt.datetime(2020, 6, 1)


@pytest.fixture(scope="function", name="timestamp_today")
def fixture_timestamp_today() -> dt.datetime:
    """Return a time stamp of todays date in date time format"""
    return dt.datetime.now()


@pytest.fixture(scope="function", name="timestamp_yesterday")
def fixture_timestamp_yesterday(timestamp_today) -> dt.datetime:
    """Return a time stamp of yesterdays date in date time format"""
    return timestamp_today - dt.timedelta(days=1)


@pytest.fixture(scope="function", name="hk_bundle_data")
def fixture_hk_bundle_data(case_id, bed_file, timestamp):
    """Get some bundle data for housekeeper"""
    data = {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [{"path": bed_file, "archive": False, "tags": ["bed", "sample"]}],
    }
    return data


@pytest.fixture(scope="function", name="sample_hk_bundle_no_files")
def fixture_sample_hk_bundle_no_files(sample, timestamp):
    """Create a complete bundle mock for testing compression"""
    hk_bundle_data = {
        "name": sample,
        "created": timestamp,
        "expires": timestamp,
        "files": [],
    }

    return hk_bundle_data


@pytest.fixture(scope="function", name="case_hk_bundle_no_files")
def fixture_case_hk_bundle_no_files(case_id, timestamp):
    """Create a complete bundle mock for testing compression"""
    hk_bundle_data = {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [],
    }

    return hk_bundle_data


@pytest.fixture(scope="function", name="compress_hk_fastq_bundle")
def fixture_compress_hk_fastq_bundle(compression_object, sample_hk_bundle_no_files):
    """Create a complete bundle mock for testing compression

    This bundle contains a pair of fastq files.
    """
    hk_bundle_data = copy.deepcopy(sample_hk_bundle_no_files)

    first_fastq = compression_object.fastq_first
    second_fastq = compression_object.fastq_second
    for fastq_file in [first_fastq, second_fastq]:
        fastq_file.touch()
        # We need to set the time to an old date
        # Create a older date
        # Convert the date to a float
        before_timestamp = dt.datetime.timestamp(dt.datetime(2020, 1, 1))
        # Update the utime so file looks old
        os.utime(fastq_file, (before_timestamp, before_timestamp))
        fastq_file_info = {"path": str(fastq_file), "archive": False, "tags": ["fastq"]}

        hk_bundle_data["files"].append(fastq_file_info)

    return hk_bundle_data


@pytest.yield_fixture(scope="function", name="housekeeper_api")
def fixture_housekeeper_api(hk_config_dict):
    """Setup Housekeeper store."""
    _api = MockHousekeeperAPI(hk_config_dict)
    yield _api


@pytest.yield_fixture(scope="function", name="real_housekeeper_api")
def fixture_real_housekeeper_api(hk_config_dict):
    """Setup a real Housekeeper store."""
    _api = HousekeeperAPI(hk_config_dict)
    _api.initialise_db()
    yield _api


@pytest.yield_fixture(scope="function", name="populated_housekeeper_api")
def fixture_populated_housekeeper_api(housekeeper_api, hk_bundle_data, helpers):
    """Setup a Housekeeper store with some data."""
    hk_api = housekeeper_api
    helpers.ensure_hk_bundle(hk_api, hk_bundle_data)
    return hk_api


@pytest.yield_fixture(scope="function", name="hk_version_obj")
def fixture_hk_version_obj(housekeeper_api, hk_bundle_data, helpers):
    """Get a housekeeper version object"""
    _version = helpers.ensure_hk_version(housekeeper_api, hk_bundle_data)
    return _version


# Process Mock


@pytest.fixture(name="process")
def fixture_process() -> ProcessMock:
    """Returns a mocked process"""
    return ProcessMock()


# Scout fixtures


@pytest.yield_fixture(scope="function", name="scout_api")
def fixture_scout_api():
    """Setup Scout api."""
    _api = MockScoutAPI()
    return _api


# Crunchy fixtures


@pytest.yield_fixture(scope="function", name="crunchy_api")
def fixture_crunchy_api():
    """Setup Crunchy api."""
    _api = MockCrunchyAPI()
    return _api


# Store fixtures


@pytest.yield_fixture(scope="function", name="analysis_store")
def fixture_analysis_store(
    base_store: Store, analysis_family: dict, wgs_application_tag: str, helpers
):
    """Setup a store instance for testing analysis API."""
    helpers.ensure_family_from_dict(
        base_store, family_info=analysis_family, app_tag=wgs_application_tag
    )

    yield base_store


@pytest.yield_fixture(scope="function", name="analysis_store_trio")
def fixture_analysis_store_trio(analysis_store):
    """Setup a store instance with a trion loaded for testing analysis API."""

    yield analysis_store


@pytest.yield_fixture(scope="function", name="analysis_store_single_case")
def fixture_analysis_store_single(base_store, analysis_family_single_case, helpers):
    """Setup a store instance with a single ind case for testing analysis API."""
    helpers.ensure_family_from_dict(base_store, family_info=analysis_family_single_case)

    yield base_store


@pytest.yield_fixture(scope="function", name="customer_group")
def fixture_customer_group() -> str:
    """Return a default customer group"""
    return "all_customers"


@pytest.yield_fixture(scope="function", name="customer_production")
def fixture_customer_production(customer_group) -> dict:
    """Return a dictionary with information about the prod customer"""
    _cust = dict(
        customer_id="cust000",
        name="Production",
        scout_access=True,
        customer_group=customer_group,
    )
    return _cust


@pytest.yield_fixture(scope="function", name="external_wgs_application_tag")
def fixture_external_wgs_application_tag() -> str:
    """Return the external wgs app tag"""
    return "WGXCUSC000"


@pytest.yield_fixture(scope="function", name="external_wgs_info")
def fixture_external_wgs_info(external_wgs_application_tag) -> dict:
    """Return a dictionary with information external WGS application"""
    _info = dict(
        application_tag=external_wgs_application_tag,
        application_type="wgs",
        description="External WGS",
        is_external=True,
        target_reads=10,
    )
    return _info


@pytest.yield_fixture(scope="function", name="external_wes_application_tag")
def fixture_external_wes_application_tag() -> str:
    """Return the external whole exome sequencing app tag"""
    return "EXXCUSR000"


@pytest.yield_fixture(scope="function", name="external_wes_info")
def fixture_external_wes_info(external_wes_application_tag) -> dict:
    """Return a dictionary with information external WES application"""
    _info = dict(
        application_tag=external_wes_application_tag,
        application_type="wes",
        description="External WES",
        is_external=True,
        target_reads=10,
    )
    return _info


@pytest.yield_fixture(scope="function", name="wgs_application_tag")
def fixture_wgs_application_tag() -> str:
    """Return the wgs app tag"""
    return "WGSPCFC060"


@pytest.yield_fixture(scope="function", name="wgs_application_info")
def fixture_wgs_application_info(wgs_application_tag) -> dict:
    """Return a dictionary with information the WGS application"""
    _info = dict(
        application_tag=wgs_application_tag,
        application_type="wgs",
        description="WGS, double",
        sequencing_depth=30,
        is_external=True,
        is_accredited=True,
        target_reads=10,
    )
    return _info


@pytest.yield_fixture(scope="function", name="store")
def fixture_store() -> Store:
    """Fixture with a CG store"""
    _store = Store(uri="sqlite:///")
    _store.create_all()
    yield _store
    _store.drop_all()


@pytest.yield_fixture(scope="function", name="base_store")
def fixture_base_store(store) -> Store:
    """Setup and example store."""
    customer_group = store.add_customer_group("all_customers", "all customers")

    store.add_commit(customer_group)
    customers = [
        store.add_customer(
            "cust000",
            "Production",
            scout_access=True,
            customer_group=customer_group,
            invoice_address="Test street",
            invoice_reference="ABCDEF",
        ),
        store.add_customer(
            "cust001",
            "Customer",
            scout_access=False,
            customer_group=customer_group,
            invoice_address="Test street",
            invoice_reference="ABCDEF",
        ),
        store.add_customer(
            "cust002",
            "Karolinska",
            scout_access=True,
            customer_group=customer_group,
            invoice_address="Test street",
            invoice_reference="ABCDEF",
        ),
        store.add_customer(
            "cust003",
            "CMMS",
            scout_access=True,
            customer_group=customer_group,
            invoice_address="Test street",
            invoice_reference="ABCDEF",
        ),
    ]
    store.add_commit(customers)
    applications = [
        store.add_application(
            tag="WGXCUSC000",
            category="wgs",
            description="External WGS",
            sequencing_depth=0,
            is_external=True,
            percent_kth=80,
            target_reads=10,
        ),
        store.add_application(
            tag="EXXCUSR000",
            category="wes",
            description="External WES",
            sequencing_depth=0,
            is_external=True,
            percent_kth=80,
            target_reads=10,
        ),
        store.add_application(
            tag="WGSPCFC060",
            category="wgs",
            description="WGS, double",
            sequencing_depth=30,
            accredited=True,
            percent_kth=80,
            target_reads=10,
        ),
        store.add_application(
            tag="RMLS05R150",
            category="rml",
            description="Ready-made",
            sequencing_depth=0,
            percent_kth=80,
            target_reads=10,
        ),
        store.add_application(
            tag="WGTPCFC030",
            category="wgs",
            description="WGS trio",
            is_accredited=True,
            sequencing_depth=30,
            target_reads=30,
            limitations="some",
            percent_kth=80,
        ),
        store.add_application(
            tag="METLIFR020",
            category="wgs",
            description="Whole genome metagenomics",
            sequencing_depth=0,
            target_reads=400000,
            percent_kth=80,
        ),
        store.add_application(
            tag="METNXTR020",
            category="wgs",
            description="Metagenomics",
            sequencing_depth=0,
            target_reads=200000,
            percent_kth=80,
        ),
        store.add_application(
            tag="MWRNXTR003",
            category="mic",
            description="Microbial whole genome ",
            sequencing_depth=0,
            percent_kth=80,
            target_reads=10,
        ),
        store.add_application(
            tag="RNAPOAR025",
            category="tgs",
            description="RNA seq, poly-A based priming",
            percent_kth=80,
            sequencing_depth=25,
            accredited=True,
            target_reads=10,
        ),
    ]

    store.add_commit(applications)

    prices = {"standard": 10, "priority": 20, "express": 30, "research": 5}
    versions = [
        store.add_version(application, 1, valid_from=dt.datetime.now(), prices=prices)
        for application in applications
    ]
    store.add_commit(versions)

    beds = [store.add_bed("Bed")]
    store.add_commit(beds)
    bed_versions = [store.add_bed_version(bed, 1, "Bed.bed") for bed in beds]
    store.add_commit(bed_versions)

    organism = store.add_organism("C. jejuni", "C. jejuni")
    store.add_commit(organism)

    yield store


@pytest.fixture(scope="function")
def sample_store(base_store) -> Store:
    """Populate store with samples."""
    new_samples = [
        base_store.add_sample("ordered", sex="male"),
        base_store.add_sample("received", sex="unknown", received=dt.datetime.now()),
        base_store.add_sample(
            "received-prepared",
            sex="unknown",
            received=dt.datetime.now(),
            prepared_at=dt.datetime.now(),
        ),
        base_store.add_sample("external", sex="female", external=True),
        base_store.add_sample(
            "external-received", sex="female", external=True, received=dt.datetime.now()
        ),
        base_store.add_sample(
            "sequenced",
            sex="male",
            received=dt.datetime.now(),
            prepared_at=dt.datetime.now(),
            sequenced_at=dt.datetime.now(),
            reads=(310 * 1000000),
        ),
        base_store.add_sample(
            "sequenced-partly",
            sex="male",
            received=dt.datetime.now(),
            prepared_at=dt.datetime.now(),
            reads=(250 * 1000000),
        ),
    ]
    customer = base_store.customers().first()
    external_app = base_store.application("WGXCUSC000").versions[0]
    wgs_app = base_store.application("WGTPCFC030").versions[0]
    for sample in new_samples:
        sample.customer = customer
        sample.application_version = external_app if "external" in sample.name else wgs_app
    base_store.add_commit(new_samples)
    return base_store


@pytest.yield_fixture(scope="function")
def disk_store(cli_runner, invoke_cli) -> Store:
    """Store on disk"""
    database = "./test_db.sqlite3"
    database_path = Path(database)
    database_uri = f"sqlite:///{database}"
    with cli_runner.isolated_filesystem():
        assert database_path.exists() is False

        # WHEN calling "init"
        result = invoke_cli(["--database", database_uri, "init"])

        # THEN it should setup the database with some tables
        assert result.exit_code == 0
        assert database_path.exists()
        assert len(Store(database_uri).engine.table_names()) > 0

        yield Store(database_uri)


@pytest.fixture(scope="function", name="trailblazer_api")
def fixture_trailblazer_api() -> MockTB:
    return MockTB()



""" This file groups all tests related to microsalt case config creation """

import logging
from pathlib import Path

from cg.apps.lims import LimsAPI
from snapshottest import Snapshot

from cg.cli.workflow.microsalt.base import config_case

EXIT_SUCCESS = 0


def test_no_arguments(cli_runner, base_context, caplog):
    """Test command without any options"""

    # GIVEN

    # WHEN dry running without anything specified
    result = cli_runner.invoke(config_case, obj=base_context)

    # THEN command should mention missing arguments
    assert result.exit_code != EXIT_SUCCESS
    assert "Aborted!" in result.output
    with caplog.at_level(logging.ERROR):
        assert "Provide ticket and/or sample" in caplog.text


def test_no_sample_found(cli_runner, base_context, caplog):
    """Test missing sample command """

    # GIVEN a not existing sample
    microbial_sample_id = "not_existing_sample"

    # WHEN dry running a sample name
    result = cli_runner.invoke(config_case, [microbial_sample_id], obj=base_context)

    # THEN command should mention missing sample
    assert result.exit_code != EXIT_SUCCESS
    with caplog.at_level(logging.ERROR):
        assert f"No sample found for that ticket/sample_id" in caplog.text


def test_no_order_found(cli_runner, base_context, caplog):
    """Test missing order command """

    # GIVEN a not existing ticket
    ticket = -1

    # WHEN dry running a order name
    result = cli_runner.invoke(config_case, ["--ticket", ticket], obj=base_context)

    # THEN command should mention missing order
    assert result.exit_code != EXIT_SUCCESS
    with caplog.at_level(logging.ERROR):
        assert f"No sample found for that ticket/sample_id" in caplog.text


def test_no_sample_order_found(cli_runner, base_context, caplog):
    """Test missing sample and order command """

    # GIVEN a not existing order
    microbial_sample_id = "not_existing_sample"
    microbial_ticket = "not_existing_order"

    # WHEN dry running a order name
    result = cli_runner.invoke(
        config_case,
        ["--ticket", microbial_ticket, microbial_sample_id],
        obj=base_context,
    )

    # THEN command should mention missing order
    assert result.exit_code != EXIT_SUCCESS
    with caplog.at_level(logging.ERROR):
        assert f"No sample found for that ticket/sample_id" in caplog.text


def test_dry_sample(
    cli_runner, base_context, microbial_sample_id, snapshot: Snapshot, lims_api: LimsAPI
):
    """Test working dry command for sample"""

    # GIVEN project, organism and reference genome is specified in lims
    lims_sample = lims_api.sample(microbial_sample_id)
    lims_sample.sample_data["project"] = {"id": "microbial_order_test"}

    # WHEN dry running a sample name
    result = cli_runner.invoke(config_case, ["--dry-run", microbial_sample_id], obj=base_context)

    # THEN command should give us a json dump
    assert result.exit_code == EXIT_SUCCESS
    snapshot.assert_match(result.output)


def test_dry_sample_order(
    cli_runner,
    base_context,
    microbial_sample_id,
    microbial_ticket,
    snapshot: Snapshot,
    lims_api: LimsAPI,
):
    """Test working dry command for sample in a order"""

    # GIVEN
    lims_sample = lims_api.sample(microbial_sample_id)
    lims_sample.sample_data["project"] = {"id": "microbial_order_test"}

    # WHEN dry running a sample name
    result = cli_runner.invoke(
        config_case,
        ["--dry-run", "--ticket", microbial_ticket, microbial_sample_id],
        obj=base_context,
    )

    # THEN command should give us a json dump
    assert result.exit_code == EXIT_SUCCESS
    snapshot.assert_match(result.output)


def test_dry_order(cli_runner, base_context, microbial_ticket, snapshot: Snapshot):
    """Test working dry command for a order"""

    # GIVEN

    # WHEN dry running a sample name
    result = cli_runner.invoke(
        config_case,
        ["--dry-run", "--ticket", microbial_ticket],
        obj=base_context,
    )

    # THEN command should give us a json dump
    assert result.exit_code == EXIT_SUCCESS
    snapshot.assert_match(result.output)


def test_sample(
    base_context, cli_runner, lims_api, microbial_sample_id, queries_path, snapshot: Snapshot
):
    """Test working command for sample"""

    # GIVEN an existing queries path
    Path(queries_path).mkdir(exist_ok=True)
    lims_api.sample(microbial_sample_id).sample_data["project"] = {"id": "microbial_order_test"}

    # WHEN dry running a sample name
    result = cli_runner.invoke(config_case, [microbial_sample_id], obj=base_context)

    # THEN command should give us a json dump
    assert result.exit_code == EXIT_SUCCESS
    outfilename = queries_path / microbial_sample_id
    outfilename = outfilename.with_suffix(".json")
    with open(outfilename, "r") as outputfile:
        snapshot.assert_match(outputfile.readlines())


def test_gonorrhoeae(cli_runner, microsalt_store, base_context, microbial_sample_id):
    """ Test if the substitution of the organism happens """
    # GIVEN a sample with organism set to gonorrhea
    sample_obj = microsalt_store.sample(microbial_sample_id)
    sample_obj.organism.internal_id = "gonorrhoeae"

    # WHEN getting the case config
    result = cli_runner.invoke(config_case, ["--dry-run", microbial_sample_id], obj=base_context)

    # THEN the organism should now be  ...
    assert "Neisseria spp." in result.output


def test_cutibacterium_acnes(cli_runner, microsalt_store, base_context, microbial_sample_id):
    """ Test if this bacteria gets its name changed """
    # GIVEN a sample with organism set to Cutibacterium acnes
    sample_obj = microsalt_store.sample(microbial_sample_id)
    sample_obj.organism.internal_id = "Cutibacterium acnes"

    # WHEN getting the case config
    result = cli_runner.invoke(config_case, ["--dry-run", microbial_sample_id], obj=base_context)

    # THEN the organism should now be ....
    assert "Propionibacterium acnes" in result.output


def test_vre_nc_017960(cli_runner, microsalt_store, base_context, microbial_sample_id):
    """ Test if this bacteria gets its name changed """
    # GIVEN a sample with organism set to VRE
    sample_obj = microsalt_store.sample(microbial_sample_id)
    sample_obj.organism.internal_id = "VRE"
    sample_obj.organism.reference_genome = "NC_017960.1"

    # WHEN getting the case config
    result = cli_runner.invoke(config_case, ["--dry-run", microbial_sample_id], obj=base_context)

    # THEN the organism should now be ....
    assert "Enterococcus faecium" in result.output


def test_vre_nc_004668(cli_runner, microsalt_store, base_context, microbial_sample_id):
    """ Test if this bacteria gets its name changed """
    # GIVEN a sample with organism set to VRE
    sample_obj = microsalt_store.sample(microbial_sample_id)
    sample_obj.organism.internal_id = "VRE"
    sample_obj.organism.reference_genome = "NC_004668.1"

    # WHEN getting the case config
    result = cli_runner.invoke(config_case, ["--dry-run", microbial_sample_id], obj=base_context)

    # THEN the organism should now be ....
    assert "Enterococcus faecalis" in result.output


def test_vre_comment(cli_runner, microsalt_store, lims_api, base_context, microbial_sample_id):
    """ Test if this bacteria gets its name changed """
    # GIVEN a sample with organism set to VRE and a comment set in LIMS
    sample_obj = microsalt_store.sample(microbial_sample_id)
    sample_obj.organism.internal_id = "VRE"
    lims_sample = lims_api.sample(microbial_sample_id)
    lims_sample.sample_data["comment"] = "ABCD123"

    # WHEN getting the case config
    result = cli_runner.invoke(config_case, ["--dry-run", microbial_sample_id], obj=base_context)

    # THEN the organism should now be ....
    assert "ABCD123" in result.output



"""Constants specific for compression"""
import datetime

# Constants for crunchy
FASTQ_FIRST_READ_SUFFIX = "_R1_001.fastq.gz"
FASTQ_SECOND_READ_SUFFIX = "_R2_001.fastq.gz"
SPRING_SUFFIX = ".spring"

# Number of days until fastqs counts as old
FASTQ_DELTA = 21
# Get the fastq delta in datetime format
FASTQ_DATETIME_DELTA = datetime.timedelta(days=FASTQ_DELTA)



"""Fixtures for cli balsamic tests"""
import json
import logging
from pathlib import Path
from datetime import datetime

import pytest

from cg.apps.hk import HousekeeperAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.gt import GenotypeAPI
from cg.meta.upload.scoutapi import UploadScoutAPI
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.store import Store
from cg.store import models
from cg.apps.scout.scout_load_config import ScoutLoadConfig

from tests.mocks.madeline import MockMadelineAPI

LOG = logging.getLogger(__name__)


@pytest.fixture(name="scout_load_config")
def fixture_scout_load_config(apps_dir) -> Path:
    """Yaml file with load information from scout"""
    return apps_dir / "scout/643594.config.yaml"


@pytest.fixture(scope="function", name="scout_hk_bundle_data")
def fixture_scout_hk_bundle_data(case_id: str, scout_load_config: Path, timestamp: datetime):
    """Get some bundle data for housekeeper"""
    tag_name = UploadScoutAPI.get_load_config_tag()

    hk_bundle_data = {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [{"path": str(scout_load_config), "archive": False, "tags": [tag_name]}],
    }
    return hk_bundle_data


@pytest.fixture(name="upload_genotypes_hk_bundle")
def fixture_upload_genotypes_hk_bundle(
    case_id: str, timestamp, case_qc_metrics: Path, bcf_file: Path
) -> dict:
    """ Returns a dictionary in hk format with files used in upload gt process"""
    data = {
        "name": case_id,
        "created": datetime.now(),
        "expires": datetime.now(),
        "files": [
            {"path": str(case_qc_metrics), "archive": False, "tags": ["qcmetrics"]},
            {"path": str(bcf_file), "archive": False, "tags": ["snv-gbcf"]},
        ],
    }
    return data


@pytest.fixture(name="analysis_obj")
def fixture_analysis_obj(
    analysis_store_trio: Store, case_id: str, timestamp: datetime, helpers
) -> models.Analysis:
    """Return a analysis object with a trio"""
    return analysis_store_trio.family(case_id).analyses[0]


@pytest.fixture(name="upload_genotypes_hk_api")
def fixture_upload_genotypes_hk_api(
    real_housekeeper_api: HousekeeperAPI,
    upload_genotypes_hk_bundle: dict,
    analysis_obj: models.Analysis,
    helpers,
) -> HousekeeperAPI:
    """Add and include files from upload genotypes hk bundle"""
    helpers.ensure_hk_bundle(real_housekeeper_api, upload_genotypes_hk_bundle)
    hk_version = real_housekeeper_api.last_version(analysis_obj.family.internal_id)
    real_housekeeper_api.include(hk_version)
    return real_housekeeper_api


@pytest.yield_fixture(name="upload_genotypes_context")
def fixture_upload_genotypes_context(
    upload_genotypes_hk_api: HousekeeperAPI, genotype_api: GenotypeAPI, analysis_store_trio: Store
) -> dict:
    """Create a upload genotypes context"""

    return {
        "genotype_api": genotype_api,
        "housekeeper_api": upload_genotypes_hk_api,
        "status_db": analysis_store_trio,
    }


@pytest.fixture(name="scout_load_object")
def fixture_scout_load_object(case_id: str, timestamp: datetime) -> ScoutLoadConfig:
    """Create a scout load config case object"""
    case_data = {
        "owner": "cust000",
        "family": case_id,
        "human_genome_build": "37",
        "rank_score_threshold": 5,
        "analysis_date": timestamp,
        "samples": [
            {"sample_id": "sample", "sex": "male", "phenotype": "affected", "analysis_type": "wgs"}
        ],
    }
    return ScoutLoadConfig(**case_data)


@pytest.fixture(scope="function", name="base_context")
def fixture_base_cli_context(
    analysis_store: Store, housekeeper_api, upload_scout_api, trailblazer_api
) -> dict:
    """context to use in cli"""
    return {
        "scout_api": MockScoutApi(),
        "scout_upload_api": upload_scout_api,
        "housekeeper_api": housekeeper_api,
        "trailblazer_api": trailblazer_api,
        "status_db": analysis_store,
        "mip-rd-dna": {"root": "hej"},
    }


@pytest.fixture(scope="function", name="vogue_context")
def fixture_vogue_cli_context(vogue_api) -> dict:
    """context to use in cli"""

    return {"vogue_api": vogue_api}


@pytest.fixture(scope="function", name="upload_scout_api")
def fixture_upload_scout_api(housekeeper_api):
    """Return a upload scout api"""
    api = MockScoutUploadApi()
    api.housekeeper = housekeeper_api

    return api


@pytest.fixture(scope="function", name="vogue_api")
def fixture_vogue_api():
    """Return a MockVogueApi"""

    return MockVogueApi()


class MockScoutApi(ScoutAPI):
    def __init__(self):
        """docstring for __init__"""
        pass

    def upload(self, scout_load_config: Path, threshold: int = 5, force: bool = False):
        """docstring for upload"""
        LOG.info("Case loaded successfully to Scout")


class MockVogueApi:
    def __init__(self):
        """docstring for __init__"""
        pass

    def load_reagent_labels(self, days: int):
        """docstring for upload"""

    def load_samples(self, days: int):
        """docstring for upload"""

    def load_flowcells(self, days: int):
        """docstring for upload"""


class MockAnalysisApi(MipAnalysisAPI):
    def __init__(self):
        """docstring for __init__"""

    def get_latest_metadata(self, internal_id):
        """docstring for upload"""
        return {}


class MockScoutUploadApi(UploadScoutAPI):
    def __init__(self, **kwargs):
        """docstring for __init__"""
        self.mock_generate_config = True
        self.housekeeper = None
        self.madeline_api = MockMadelineAPI()
        self.analysis = MockAnalysisApi()
        self.config = {}
        self.file_exists = False
        self.lims = MockLims()

    @pytest.fixture(autouse=True)
    def _request_analysis(self, analysis_store_single_case):
        self.analysis = analysis_store_single_case

    def generate_config(self, analysis_obj, **kwargs):
        """Mock the generate config"""
        if self.mock_generate_config:
            return self.config

        return super().generate_config(analysis_obj, **kwargs)

    def save_config_file(self, scout_config, file_path):
        """docstring for save_config_file"""
        return

    def add_scout_config_to_hk(self, config_file_path: str, case_id: str, delete: bool = False):
        """docstring for add_scout_config_to_hk"""
        LOG.info("Use mock to upload file")
        if self.file_exists:
            raise FileExistsError("Scout config already exists")


class MockLims:
    """Mock lims fixture"""

    lims = None

    def __init__(self):
        self.lims = self

    @staticmethod
    def lims_samples():
        """ Return LIMS-like family samples """
        lims_family = json.load(open("tests/fixtures/report/lims_family.json"))
        return lims_family["samples"]

    def sample(self, sample_id):
        """ Returns a lims sample matching the provided sample_id """
        for sample in self.lims_samples():
            if sample["id"] == sample_id:
                return sample
        return None



"""This script tests the cli method enable workflows"""
from cg.cli.workflow.base import workflow as workflow_cmd

EXIT_SUCCESS = 0


def test_no_options(cli_runner, base_context):
    """Test command with no options"""

    # GIVEN

    # WHEN dry running
    result = cli_runner.invoke(workflow_cmd, obj=base_context)

    # THEN command should have returned all pipelines that is supported
    assert result.exit_code == EXIT_SUCCESS
    assert "balsamic" in result.output
    assert "microsalt" in result.output
    assert "mip-dna" in result.output
    assert "mip-rna" in result.output



"""This script tests the cli methods to add families to status-db"""
from datetime import datetime, timedelta

from cg.constants import Pipeline
from cg.store import Store


def test_that_many_families_can_have_one_sample_each(base_store: Store):
    """Test that tests that families are returned even if there are many result rows in the query"""

    # GIVEN a database with two families one with 50 sequenced samples
    # the other family with one
    n_test_families = 50
    test_families = add_families_with_samples(base_store, n_test_families, sequenced=True)

    # WHEN getting families to analyse
    families = base_store.cases_to_analyze(pipeline=Pipeline.MIP_DNA)

    # THEN families should contain the test family
    assert len(families) == len(test_families)


def test_that_families_can_have_many_samples(base_store: Store):
    """Test that tests that families are returned even if there are many result rows in the query"""

    # GIVEN a database with two families one with 50 sequenced samples
    # the other family with one
    default_limit = 50
    test_family_50 = add_family_with_samples(
        base_store, "test_family_50åäp+ölo0l", default_limit, sequenced=True
    )
    test_sample = add_sample(base_store, sequenced=True)
    test_family = add_family(base_store, "family_one_sample")
    base_store.relate_sample(test_family, test_sample, "unknown")

    # WHEN getting families to analyse
    families = base_store.cases_to_analyze(pipeline=Pipeline.MIP_DNA)

    # THEN families should contain the test family
    assert families
    assert test_family_50 in families
    assert test_family in families


def test_external_sample_to_re_analyse(base_store: Store):
    """Test that a family marked for re-analyse with one sample external not sequenced inhouse and
    with completed analysis do show up among the families to analyse"""

    # GIVEN a database with a family with one not sequenced external sample and completed analysis
    test_sample = add_sample(base_store, sequenced=False, external=True)
    test_analysis = add_analysis(base_store, completed=True, reanalyse=True)
    base_store.relate_sample(test_analysis.family, test_sample, "unknown")

    # WHEN getting families to analyse
    families = base_store.cases_to_analyze(pipeline=Pipeline.MIP_DNA)

    # THEN families should contain the test family
    assert families
    assert test_analysis.family in families


def test_family_to_re_analyse(base_store: Store):
    """Test that a family marked for re-analyse with one sample that has been sequenced and
    with completed analysis do show up among the families to analyse"""

    # GIVEN a database with a family with one of one sequenced samples and completed analysis
    test_sample = add_sample(base_store, sequenced=True)
    test_analysis = add_analysis(base_store, completed=True, reanalyse=True)
    base_store.relate_sample(test_analysis.family, test_sample, "unknown")

    # WHEN getting families to analyse
    families = base_store.cases_to_analyze(pipeline=Pipeline.MIP_DNA)

    # THEN families should contain the test family
    assert families
    assert test_analysis.family in families


def test_all_samples_and_analysis_completed(base_store: Store):
    """Test that a family with one sample that has been sequenced and with completed
    analysis don't show up among the families to analyse"""

    # GIVEN a database with a family with one of one sequenced samples and completed analysis
    test_sample = add_sample(base_store, sequenced=True)
    test_analysis = add_analysis(base_store, completed=True)
    base_store.relate_sample(test_analysis.family, test_sample, "unknown")

    # WHEN getting families to analyse
    families = base_store.cases_to_analyze(pipeline=Pipeline.MIP_DNA)

    # THEN families should not contain the test family

    assert not families


def test_specified_analysis_in_result(base_store: Store):
    """Test that a family with one sample that has specified data_analysis does show up"""

    # GIVEN a database with a family with one sequenced samples for MIP analysis
    pipeline = Pipeline.BALSAMIC
    test_sample = add_sample(base_store, sequenced=True)
    test_family = add_family(base_store, data_analysis=pipeline)
    base_store.relate_sample(test_family, test_sample, "unknown")

    # WHEN getting families to analyse
    families = base_store.cases_to_analyze(pipeline=pipeline)

    # THEN families should contain the test family
    assert families
    assert test_family in families


def test_exclude_other_pipeline_analysis_from_result(base_store: Store):
    """Test that a family with specified analysis and with one sample does not show up among
    others"""

    # GIVEN a database with a family with one sequenced samples for specified analysis
    test_sample = add_sample(base_store, sequenced=True)
    test_family = add_family(base_store, data_analysis=Pipeline.BALSAMIC)
    base_store.relate_sample(test_family, test_sample, "unknown")

    # WHEN getting families to analyse
    families = base_store.cases_to_analyze(pipeline=Pipeline.MIP_DNA)

    # THEN families should not contain the test family
    assert not families


def test_one_of_two_sequenced_samples(base_store: Store):
    """Test that a family with one of one samples that has been sequenced shows up among the
    families to analyse"""

    # GIVEN a database with a family with one of one sequenced samples and no analysis
    test_family = add_family(base_store)
    test_sample1 = add_sample(base_store, sequenced=True)
    test_sample2 = add_sample(base_store, sequenced=False)
    base_store.relate_sample(test_family, test_sample1, "unknown")
    base_store.relate_sample(test_family, test_sample2, "unknown")

    # WHEN getting families to analyse
    families = base_store.cases_to_analyze(pipeline=Pipeline.MIP_DNA)

    # THEN families should not contain the test family
    assert not families


def test_one_of_one_sequenced_samples(base_store: Store):
    """Test that a family with one of one samples that has been sequenced shows up among the
    families to analyse"""

    # GIVEN a database with a family with one of one sequenced samples and no analysis
    test_family = add_family(base_store)
    test_sample = add_sample(base_store, sequenced=True)
    base_store.relate_sample(test_family, test_sample, "unknown")
    assert test_sample.sequenced_at is not None

    # WHEN getting families to analyse
    families = base_store.cases_to_analyze(pipeline=Pipeline.MIP_DNA)

    # THEN families should contain the test family
    assert families
    assert test_family in families


def ensure_application_version(disk_store, application_tag="dummy_tag"):
    """utility function to return existing or create application version for tests"""
    application = disk_store.application(tag=application_tag)
    if not application:
        application = disk_store.add_application(
            tag=application_tag,
            category="wgs",
            description="dummy_description",
            percent_kth=80,
        )
        disk_store.add_commit(application)

    prices = {"standard": 10, "priority": 20, "express": 30, "research": 5}
    version = disk_store.application_version(application, 1)
    if not version:
        version = disk_store.add_version(application, 1, valid_from=datetime.now(), prices=prices)
        disk_store.add_commit(version)
    return version


def ensure_customer(disk_store, customer_id="cust_test"):
    """utility function to return existing or create customer for tests"""
    customer_group = disk_store.customer_group("dummy_group")
    if not customer_group:
        customer_group = disk_store.add_customer_group("dummy_group", "dummy group")

        customer = disk_store.add_customer(
            internal_id=customer_id,
            name="Test Customer",
            scout_access=False,
            customer_group=customer_group,
            invoice_address="dummy_address",
            invoice_reference="dummy_reference",
        )
        disk_store.add_commit(customer)
    customer = disk_store.customer(customer_id)
    return customer


def add_sample(
    store,
    sample_name="sample_test",
    received=False,
    prepared=False,
    sequenced=False,
    delivered=False,
    invoiced=False,
    external=None,
):
    """utility function to add a sample to use in tests"""
    customer = ensure_customer(store)
    application_version_id = ensure_application_version(store).id
    sample = store.add_sample(name=sample_name, sex="unknown")
    sample.application_version_id = application_version_id
    sample.customer = customer
    if received:
        sample.received_at = datetime.now()
    if prepared:
        sample.prepared_at = datetime.now()
    if sequenced:
        sample.sequenced_at = datetime.now()
    if delivered:
        sample.delivered_at = datetime.now()
    if invoiced:
        invoice = store.add_invoice(customer)
        sample.invoice = invoice
    if external:
        sample.is_external = external

    store.add_commit(sample)
    return sample


def add_samples(base_store, n_samples, sequenced):
    """utility function to add many samples to use in tests"""
    samples = []
    for _ in range(n_samples):
        samples.append(add_sample(base_store, sequenced=sequenced))
    return samples


def relate_samples(base_store, family, samples):
    """utility function to relate many samples to one family"""

    for sample in samples:
        base_store.relate_sample(family, sample, "unknown")


def add_family_with_samples(base_store, family_name, n_samples, sequenced):
    """utility function to add one family with many samples to use in tests"""

    test_samples = add_samples(base_store, n_samples, sequenced=sequenced)
    test_family = add_family(base_store, family_name)
    relate_samples(base_store, test_family, test_samples)
    return test_family


def add_families_with_samples(base_store, n_families, sequenced):
    """utility function to add many families with two samples to use in tests"""

    families = []
    for i in range(n_families):
        family = add_family_with_samples(base_store, f"f{i}", 2, sequenced=sequenced)
        families.append(family)
    return families


def ensure_panel(disk_store, panel_id="panel_test", customer_id="cust_test"):
    """utility function to add a panel to use in tests"""
    customer = ensure_customer(disk_store, customer_id)
    panel = disk_store.panel(panel_id)
    if not panel:
        panel = disk_store.add_panel(
            customer=customer,
            name=panel_id,
            abbrev=panel_id,
            version=1.0,
            date=datetime.now(),
            genes=1,
        )
        disk_store.add_commit(panel)
    return panel


def add_family(
    disk_store,
    family_id="family_test",
    customer_id="cust_test",
    ordered_days_ago=0,
    action=None,
    priority=None,
    data_analysis=Pipeline.MIP_DNA,
):
    """utility function to add a family to use in tests"""
    panel = ensure_panel(disk_store)
    customer = ensure_customer(disk_store, customer_id)
    family = disk_store.add_family(data_analysis=data_analysis, name=family_id, panels=panel.name)
    family.customer = customer
    family.ordered_at = datetime.now() - timedelta(days=ordered_days_ago)
    if action:
        family.action = action
    if priority:
        family.priority = priority
    disk_store.add_commit(family)
    return family


def add_analysis(
    store, completed=False, uploaded=False, pipeline=Pipeline.BALSAMIC, reanalyse=False
):
    """Utility function to add an analysis for tests"""
    analysis = store.add_analysis(pipeline=pipeline, version="")
    if completed:
        analysis.completed_at = datetime.now()
    if uploaded:
        analysis.uploaded_at = datetime.now()

    family = add_family(store)

    family.analyses.append(analysis)
    store.add_commit(analysis)

    if reanalyse:
        family.action = "analyze"

    store.add_commit(family)

    return analysis



"""Tests for the models in scout load config"""

from cg.apps.scout import scout_load_config


def test_validate_individual_display_name():
    """Test to validate an individual"""
    # GIVEN some sample information
    sample = {
        "analysis_type": "wgs",
        "bam_path": "/path/to/sample.bam",
        "capture_kit": None,
        "father": "0",
        "mother": "0",
        "sample_id": "sample_id",
        "sample_name": "sample_name",
        "sex": "male",
        "tissue_type": "unknown",
        "phenotype": "affected",
    }

    # WHEN validating the sample data
    ind_obj = scout_load_config.ScoutIndividual(**sample)

    # THEN assert that the display name is correct
    assert ind_obj.sample_name == sample["sample_name"]



"""Tests for madeline extension"""

from pathlib import Path

from cg.apps.madeline.api import MadelineAPI


def get_ind_info(columns, line):
    """Take some madeline columns and a raw madeline line and create a dictionary"""
    ind_info = dict(zip(columns, line.rstrip().split("\t")))
    return ind_info


def test_run_madeline(trio: dict, populated_madeline_api: MadelineAPI, project_dir: Path):
    """Test to run the madeline call from api"""
    # GIVEN a populated madeline api
    madeline_api = populated_madeline_api
    # GIVEN a outfile
    outfile = project_dir / "madeline.xml"
    outfile.touch()

    # WHEN running the madeline command
    out_path: Path = madeline_api.run("a family", trio, str(outfile))
    # THEN assert a madeline xml file is returned
    assert out_path.suffix == ".xml"


def test_generate_madeline_input_no_mother(madeline_columns: dict, proband: dict):
    """Test generate input for madeline when mother is missing"""
    # GIVEN a family id and a ind with unknown mother
    family_id = "test"
    proband.pop("mother")
    inds = [proband]
    # WHEN generating madeline formated lines
    madeline_lines = MadelineAPI.make_ped(family_id, inds)
    nr_lines = 0
    for nr_lines, line in enumerate(madeline_lines, 1):
        if nr_lines == 1:
            continue
        ind_info = get_ind_info(madeline_columns.keys(), line)

    # THEN assert that the mother is set to '.', meaning no mother exists
    assert ind_info["mother"] == "."


def test_generate_madeline_input_no_sex(madeline_columns, proband):
    """Test generate input for madeline when sex is missing"""

    # GIVEN a family id and a ind with unknown sex
    family_id = "test"
    proband["sex"] = "unknown"
    inds = [proband]
    # WHEN generating madeline formated lines
    madeline_lines = MadelineAPI.make_ped(family_id, inds)
    i = 0
    for i, line in enumerate(madeline_lines, 1):
        if i == 1:
            continue
        ind_info = get_ind_info(madeline_columns.keys(), line)

    # THEN assert that sex is set to '.', meaning sex is unknown
    assert ind_info["sex"] == "."


def test_generate_madeline_input(madeline_columns, proband):
    """Test generate input for madeline"""

    # GIVEN a family id and a list of ind dicts
    family_id = "test"
    inds = [proband]
    # WHEN generating madeline formated lines
    madeline_lines = MadelineAPI.make_ped(family_id, inds)
    # Skip the header line
    next(madeline_lines)
    # Convert line to dict
    ind_info = get_ind_info(madeline_columns.keys(), next(madeline_lines))

    # THEN assert that the family id is included
    assert ind_info["family"] == family_id
    # THEN assert that the ind id information is correct
    assert ind_info["sample"] == proband["sample"]
    # THEN assert that the is converted to madeline format
    assert ind_info["sex"] == "F"


def test_generate_madeline_input_none_status(madeline_columns, proband):
    """Test generate input for madeline"""

    # GIVEN a family id and a list of ind dicts
    proband["status"] = None
    family_id = "test"
    inds = [proband]
    # WHEN generating madeline formated lines
    madeline_lines = MadelineAPI.make_ped(family_id, inds)
    # Skip the header line
    next(madeline_lines)
    # Convert line to dict
    ind_info = get_ind_info(madeline_columns.keys(), next(madeline_lines))

    # THEN assert that the status is "."
    assert ind_info["status"] == "."


def test_generate_madeline_input_non_existing_status(madeline_columns, proband):
    """Test generate input for madeline"""

    # GIVEN an individual without status
    proband.pop("status")
    assert "status" not in proband
    family_id = "test"
    # GIVEN a family id and a list of ind dicts
    inds = [proband]
    # WHEN generating madeline formated lines
    madeline_lines = MadelineAPI.make_ped(family_id, inds)
    # Skip the header line
    next(madeline_lines)
    # Convert line to dict
    ind_info = get_ind_info(madeline_columns.keys(), next(madeline_lines))

    # THEN assert that the status is "."
    assert ind_info["status"] == "."


def test_generate_madeline_input_no_inds(madeline_columns):
    """Test generate input for madeline when no individuals"""

    # GIVEN a family id and a empty list of inds
    family_id = "test"
    inds = []
    # WHEN generating madeline formated lines
    res = MadelineAPI.make_ped(family_id, inds)
    # THEN assert that only the header line was generated
    i = 0
    header_line = None
    for i, line in enumerate(res, 1):
        header_line = line
    assert header_line == "\t".join(madeline_columns.values())
    # THEN assert only the header line is returned
    assert i == 1



"""Test methods for cg/cli/set/samples"""
import logging

import pytest

from cg.cli.set.families import families
from cg.store import Store

SUCCESS = 0


@pytest.mark.parametrize("identifier_key", ["ticket_number", "order"])
def test_set_families_by_sample_identifiers(
    cli_runner, base_context, base_store: Store, identifier_key, helpers, caplog
):
    # GIVEN a database with a family with a sample
    sample_obj = helpers.add_sample(base_store)
    sample_obj.ticket_number = 123456
    sample_obj.order = "An order"
    case = helpers.add_family(base_store)
    helpers.add_relationship(base_store, sample=sample_obj, family=case)
    identifier_value = getattr(sample_obj, identifier_key)

    caplog.set_level(logging.INFO)

    # WHEN calling set families with valid sample identifiers
    result = cli_runner.invoke(
        families,
        ["--sample-identifier", identifier_key, identifier_value],
        obj=base_context,
        catch_exceptions=False,
    )

    # THEN it should name the family to be changed
    assert case.internal_id in caplog.text
    assert case.name in caplog.text



"""Test how the api handles files"""
import logging
from pathlib import Path

from cg.apps.hk import HousekeeperAPI


def test_new_file(housekeeper_api, bed_file, small_helpers):
    """Test to create a new file with the housekeeper api"""
    # GIVEN a housekeeper api without files and the path to an existing file
    assert small_helpers.length_of_iterable(housekeeper_api.files()) == 0
    assert Path(bed_file).exists() is True

    # WHEN creating a new file
    new_file_obj = housekeeper_api.new_file(path=bed_file)

    # THEN assert a file object was created
    assert new_file_obj
    # THEN assert that the path is correct
    assert new_file_obj.path == bed_file
    # THEN assert that no file is added to the database
    assert small_helpers.length_of_iterable(housekeeper_api.files()) == 0


def test_new_file_non_existing_path(housekeeper_api):
    """Test to create a new file with the housekeeper api"""
    # GIVEN a housekeeper api without files and the path to a not existing file
    file_name = "a_file.hello"
    assert Path(file_name).exists() is False

    # WHEN creating a new file
    new_file_obj = housekeeper_api.new_file(path=file_name)

    # THEN assert a file object was created
    assert new_file_obj
    # THEN assert that the path is correct
    assert new_file_obj.path == file_name


def test_add_new_file(populated_housekeeper_api, case_id, madeline_output, small_helpers):
    """Test to create a new file with the housekeeper api"""
    # GIVEN a populated housekeeper api and the existing version of a bundle
    version_obj = populated_housekeeper_api.last_version(bundle=case_id)
    # GIVEN an existing file that is not included in the database
    assert Path(madeline_output).exists() is True
    # GIVEN a tag that does not exist
    tag = "madeline"
    assert populated_housekeeper_api.tag(name=tag) is None
    # GIVEN a known number of files in the db
    nr_files_in_db = small_helpers.length_of_iterable(populated_housekeeper_api.files())

    # WHEN creating a new file
    new_file_obj = populated_housekeeper_api.add_file(
        path=madeline_output, version_obj=version_obj, tags=tag
    )

    # THEN assert a file object was created
    assert new_file_obj
    # THEN assert that the path is correct
    assert new_file_obj.path == str(Path(madeline_output).resolve())
    # THEN assert that the file was not added to the database
    new_nr_files = small_helpers.length_of_iterable(populated_housekeeper_api.files())
    assert new_nr_files == nr_files_in_db + 1


def test_get_files(populated_housekeeper_api, case_id, tags, small_helpers):
    """Test to fetch files with the get files method"""
    # GIVEN a populated housekeeper api with some files
    nr_files = small_helpers.length_of_iterable(populated_housekeeper_api.files())
    assert nr_files > 0

    # WHEN fetching all files
    files = populated_housekeeper_api.get_files(bundle=case_id, tags=tags)

    # THEN assert all files where fetched
    assert small_helpers.length_of_iterable(files) == nr_files


def test_get_file(populated_housekeeper_api: HousekeeperAPI):
    """Test to fetch a file from the database"""
    # GIVEN a housekeeper api with a file
    file_obj = populated_housekeeper_api.files().first()
    assert file_obj

    # GIVEN the id of a file that exists in HK
    file_id = file_obj.id

    # WHEN fetching the file with get_file
    hk_file = populated_housekeeper_api.get_file(file_id)

    # THEN assert a file was returned
    assert hk_file is not None


def test_delete_file(populated_housekeeper_api: HousekeeperAPI):
    """Test to fetch a file from the database"""
    # GIVEN a housekeeper api with a file
    file_obj = populated_housekeeper_api.files().first()
    assert file_obj

    # GIVEN the id of a file that exists in HK
    file_id = file_obj.id

    # WHEN deleting the file
    populated_housekeeper_api.delete_file(file_id)

    # THEN assert a file was removed
    assert populated_housekeeper_api.get_file(file_id) is None


def test_get_included_path(populated_housekeeper_api, case_id):
    """Test to get the included path for a file"""
    # GIVEN a populated housekeeper api and the root dir
    root_dir = Path(populated_housekeeper_api.get_root_dir())
    # GIVEN a version and a file object
    version_obj = populated_housekeeper_api.last_version(case_id)
    file_obj = version_obj.files[0]

    # WHEN fetching the included path
    included_path = populated_housekeeper_api.get_included_path(
        root_dir=root_dir, version_obj=version_obj, file_obj=file_obj
    )

    # THEN assert that there is no file existing in the included path
    assert included_path.exists() is False
    # THEN assert that the correct path was created
    assert included_path == root_dir / version_obj.relative_root_dir / Path(file_obj.path).name


def test_get_include_file(populated_housekeeper_api, case_id):
    """Test to get the included path for a file"""
    # GIVEN a populated housekeeper api and the root dir
    root_dir = Path(populated_housekeeper_api.get_root_dir())
    version_obj = populated_housekeeper_api.last_version(case_id)
    file_obj = version_obj.files[0]
    original_path = Path(file_obj.path)
    included_path = root_dir / version_obj.relative_root_dir / original_path.name
    # GIVEN that the included file does not exist
    assert included_path.exists() is False

    # WHEN including the file
    included_file = populated_housekeeper_api.include_file(file_obj, version_obj)

    # THEN assert that the file has been linked to the included place
    assert included_path.exists() is True
    # THEN assert that the file path has been updated
    assert included_file.path != original_path



"""Tests for delivery API"""

from pathlib import Path
from typing import List, Set
from cg.store.models import Sample, FamilySample
from cg.meta.deliver import DeliverAPI
from cg.store import Store
from cg.apps.hk import HousekeeperAPI

from housekeeper.store import models as hk_models

from tests.store_helpers import StoreHelpers


def test_get_delivery_path(
    base_store: Store, real_housekeeper_api: HousekeeperAPI, project_dir: Path, case_id: str
):
    """Test to create the delivery path"""
    # GIVEN a deliver api
    deliver_api = DeliverAPI(
        store=base_store,
        hk_api=real_housekeeper_api,
        case_tags=["case-tag"],
        sample_tags=["sample-tag"],
        project_base_path=project_dir,
    )
    customer_id = "cust000"
    ticket_id = 1234
    deliver_api._set_customer_id(customer_id)
    deliver_api._set_ticket_id(ticket_id)

    # WHEN fetching the deliver path
    deliver_path = deliver_api.create_delivery_dir_path(case_name=case_id)

    # THEN assert that the path looks like expected
    assert deliver_path == project_dir / customer_id / "inbox" / str(ticket_id) / case_id


def test_get_case_analysis_files(populated_deliver_api: DeliverAPI, case_id: str):
    """Test to fetch case specific files for a case that exists in housekeeper"""
    deliver_api: DeliverAPI = populated_deliver_api
    # GIVEN a case which exist as bundle in hk with a version
    version_obj = deliver_api.hk_api.last_version(case_id)
    assert version_obj

    # GIVEN that a case object exists in the database
    link_objs: List[FamilySample] = deliver_api.store.family_samples(case_id)
    samples: List[Sample] = [link.sample for link in link_objs]
    sample_ids: Set[str] = set([sample.internal_id for sample in samples])

    # WHEN fetching all case files from the delivery api
    bundle_latest_files = deliver_api.get_case_files_from_version(
        version_obj=version_obj, sample_ids=sample_ids
    )

    ## THEN housekeeper files should be returned
    assert bundle_latest_files


def test_get_case_files_from_version(
    analysis_store: Store,
    case_id: str,
    real_housekeeper_api: HousekeeperAPI,
    case_hk_bundle_no_files: dict,
    bed_file: str,
    vcf_file: Path,
    project_dir: Path,
    helpers=StoreHelpers,
):
    # GIVEN a store with a case
    case_obj = analysis_store.family(case_id)
    assert case_obj.internal_id == case_id
    # GIVEN a delivery api
    deliver_api = DeliverAPI(
        store=analysis_store,
        hk_api=real_housekeeper_api,
        case_tags=[{"case-tag"}],
        sample_tags=[{"sample-tag"}],
        project_base_path=project_dir,
    )

    # GIVEN a housekeeper db populated with a bundle including a case specific file and a sample specific file
    case_hk_bundle_no_files["files"] = [
        {"path": bed_file, "archive": False, "tags": ["case-tag"]},
        {"path": str(vcf_file), "archive": False, "tags": ["sample-tag", "ADM1"]},
    ]
    helpers.ensure_hk_bundle(real_housekeeper_api, bundle_data=case_hk_bundle_no_files)

    # GIVEN a version object where two file exists
    version_obj: hk_models.Version = real_housekeeper_api.last_version(case_id)
    assert len(version_obj.files) == 2

    # GIVEN the sample ids of the samples
    link_objs: List[FamilySample] = analysis_store.family_samples(case_id)
    samples: List[Sample] = [link.sample for link in link_objs]
    sample_ids: Set[str] = set([sample.internal_id for sample in samples])

    # WHEN fetching the case files
    case_files = deliver_api.get_case_files_from_version(
        version_obj=version_obj, sample_ids=sample_ids
    )

    # THEN we should only get the case specific files back
    nr_files: int = 0
    case_file: Path
    for nr_files, case_file in enumerate(case_files, 1):
        assert case_file.name == Path(bed_file).name
    # THEN assert that only the case-tag file was returned
    assert nr_files == 1


def test_get_sample_files_from_version(
    deliver_api: DeliverAPI,
    case_hk_bundle_no_files: dict,
    bed_file: str,
    vcf_file: Path,
    project_dir: Path,
    case_id: str,
    helpers=StoreHelpers,
):
    """Test to fetch sample specific files from the deliver API

    The purpose of the test is to see that only sample specific files are returned
    """
    # GIVEN a case which exist as bundle in hk
    # GIVEN a housekeeper db populated with a bundle including a case specific file and a sample specific file
    hk_api = deliver_api.hk_api
    case_hk_bundle_no_files["files"] = [
        {"path": bed_file, "archive": False, "tags": ["case-tag"]},
        {"path": str(vcf_file), "archive": False, "tags": ["sample-tag", "ADM1"]},
    ]
    helpers.ensure_hk_bundle(hk_api, bundle_data=case_hk_bundle_no_files)
    # GIVEN a version object with some files
    version_obj: hk_models.Version = hk_api.last_version(case_id)
    assert len(version_obj.files) == 2

    # WHEN fetching the sample specific files
    sample_files = deliver_api.get_sample_files_from_version(
        version_obj=version_obj, sample_id="ADM1"
    )

    # THEN assert that only the sample specific file was returned
    nr_files: int = 0
    sample_file: Path
    for nr_files, sample_file in enumerate(sample_files, 1):
        assert sample_file.name == vcf_file.name
    # THEN assert that only the sample-tag file was returned
    assert nr_files == 1



"""
    Module for interacting with crunchy to perform:
        1. Compressing: FASTQ to SPRING
        2. Decompressing: SPRING to FASTQ
    along with the helper methods
"""

import datetime
import json
import logging
import tempfile
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import List, Optional

from marshmallow import ValidationError

from cg.constants import FASTQ_DELTA
from cg.models import CompressionData
from cg.utils import Process
from cg.utils.date import get_date_str

from .models import CrunchyFileSchema
from .sbatch import SBATCH_FASTQ_TO_SPRING, SBATCH_HEADER_TEMPLATE, SBATCH_SPRING_TO_FASTQ

LOG = logging.getLogger(__name__)


FLAG_PATH_SUFFIX = ".crunchy.txt"
PENDING_PATH_SUFFIX = ".crunchy.pending.txt"


class CrunchyAPI:
    """
    API for crunchy
    """

    def __init__(self, config: dict):

        self.process = Process("sbatch")
        self.slurm_account = config["crunchy"]["slurm"]["account"]
        self.crunchy_env = config["crunchy"]["slurm"]["conda_env"]
        self.mail_user = config["crunchy"]["slurm"]["mail_user"]
        self.reference_path = config["crunchy"]["cram_reference"]
        self.dry_run = False

    def set_dry_run(self, dry_run: bool) -> None:
        """Update dry run"""
        LOG.info("Updating compress api")
        LOG.info("Set dry run to %s", dry_run)
        self.dry_run = dry_run

    # Methods to check compression status
    @staticmethod
    def is_compression_pending(compression_obj: CompressionData) -> bool:
        """Check if compression/decompression has started but not finished"""
        if compression_obj.pending_exists():
            LOG.info("Compression/decompression is pending for %s", compression_obj.run_name)
            return True
        LOG.info("Compression/decompression is not running")
        return False

    def is_fastq_compression_possible(self, compression_obj: CompressionData) -> bool:
        """Check if FASTQ compression is possible

        There are three possible answers to this question:

         - Compression is running          -> Compression NOT possible
         - SPRING archive exists           -> Compression NOT possible
         - Not compressed and not running  -> Compression IS possible
        """
        if self.is_compression_pending(compression_obj):
            return False

        if compression_obj.spring_exists():
            LOG.info("SPRING file found")
            return False

        LOG.info("FASTQ compression is possible")

        return True

    @staticmethod
    def is_spring_decompression_possible(compression_obj: CompressionData) -> bool:
        """Check if SPRING decompression is possible

        There are three possible answers to this question:

            - Compression/Decompression is running      -> Decompression is NOT possible
            - The FASTQ files are not compressed        -> Decompression is NOT possible
            - Compression has been performed            -> Decompression IS possible

        """
        if compression_obj.pending_exists():
            LOG.info("Compression/decompression is pending for %s", compression_obj.run_name)
            return False

        if not compression_obj.spring_exists():
            LOG.info("No SPRING file found")
            return False

        if compression_obj.pair_exists():
            LOG.info("FASTQ files already exists")
            return False

        LOG.info("Decompression is possible")

        return True

    def is_fastq_compression_done(self, compression_obj: CompressionData) -> bool:
        """Check if FASTQ compression is finished

        This is checked by controlling that the SPRING files that are produced after FASTQ
        compression exists.

        The following has to be fulfilled for FASTQ compression to be considered done:

            - A SPRING archive file exists
            - A SPRING archive metada file exists
            - The SPRING archive has not been unpacked before FASTQ delta (21 days)

        Note:
        'updated_at' indicates at what date the SPRING archive was unarchived last.
        If the SPRING archive has never been unarchived 'updated_at' is None

        """
        LOG.info("Check if FASTQ compression is finished")
        LOG.info("Check if SPRING file %s exists", compression_obj.spring_path)
        if not compression_obj.spring_exists():
            LOG.info("No SPRING file for %s", compression_obj.run_name)
            return False
        LOG.info("SPRING file found")

        LOG.info("Check if SPRING metadata file %s exists", compression_obj.spring_metadata_path)
        if not compression_obj.metadata_exists():
            LOG.info("No metadata file found")
            return False
        LOG.info("SPRING metadata file found")

        spring_metadata = self.get_spring_metadata(compression_obj.spring_metadata_path)
        # Check if the SPRING archive has been unarchived
        updated_at = self.get_file_updated_at(spring_metadata)

        if updated_at is None:
            LOG.info("FASTQ compression is done for %s", compression_obj.run_name)
            return True

        LOG.info("Files where unpacked %s", updated_at)

        if not self.check_if_update_spring(updated_at):
            return False

        LOG.info("FASTQ compression is done for %s", compression_obj.run_name)

        return True

    def is_spring_decompression_done(self, compression_obj: CompressionData) -> bool:
        """Check if SPRING decompression if finished.

        This means that all three files specified in SPRING metadata should exist.
        That is

            - First read in FASTQ pair should exist
            - Second read in FASTQ pair should exist
            - SPRING archive file should still exist
        """

        spring_metadata_path = compression_obj.spring_metadata_path
        LOG.info("Check if SPRING metadata file %s exists", spring_metadata_path)

        if not compression_obj.metadata_exists():
            LOG.info("No SPRING metadata file found")
            return False

        spring_metadata = self.get_spring_metadata(spring_metadata_path)
        if not spring_metadata:
            LOG.info("Malformed metadata content")
            return False

        for file_info in spring_metadata:
            if not Path(file_info["path"]).exists():
                LOG.info("File %s does not exist", file_info["path"])
                return False
            if "updated" not in file_info:
                LOG.info("Files have not been unarchived")
                return False

        LOG.info("SPRING decompression is done for run %s", compression_obj.run_name)

        return True

    # These are the compression/decompression methods
    def fastq_to_spring(self, compression_obj: CompressionData, sample_id: str = "") -> None:
        """
        Compress FASTQ files into SPRING by sending to sbatch SLURM

        """
        job_name = "_".join([sample_id, compression_obj.run_name, "fastq_to_spring"])

        log_dir = self.get_log_dir(compression_obj.spring_path)
        sbatch_header = self._get_slurm_header(job_name=job_name, log_dir=str(log_dir))

        sbatch_body = self._get_slurm_fastq_to_spring(compression_obj=compression_obj)
        sbatch_path = self.get_sbatch_path(log_dir, "fastq", compression_obj.run_name)
        sbatch_content = "\n".join([sbatch_header, sbatch_body])
        self._submit_sbatch(
            sbatch_content=sbatch_content,
            sbatch_path=sbatch_path,
            pending_path=compression_obj.pending_path,
        )

    def spring_to_fastq(self, compression_obj: CompressionData, sample_id: str = "") -> None:
        """
        Decompress SPRING into FASTQ by submitting sbatch script to SLURM

        """
        spring_metadata = self.get_spring_metadata(compression_obj.spring_metadata_path)

        job_name = "_".join([sample_id, compression_obj.run_name, "spring_to_fastq"])
        log_dir = self.get_log_dir(compression_obj.spring_path)

        sbatch_header = self._get_slurm_header(job_name=job_name, log_dir=str(log_dir))

        files_info = self.get_spring_archive_files(spring_metadata)

        sbatch_body = self._get_slurm_spring_to_fastq(
            compression_obj=compression_obj,
            checksum_first=files_info["fastq_first"]["checksum"],
            checksum_second=files_info["fastq_second"]["checksum"],
        )

        sbatch_path = self.get_sbatch_path(log_dir, "spring", compression_obj.run_name)
        sbatch_content = "\n".join([sbatch_header, sbatch_body])
        self._submit_sbatch(
            sbatch_content=sbatch_content,
            sbatch_path=sbatch_path,
            pending_path=compression_obj.pending_path,
        )

    # Spring metadata methods
    def get_spring_metadata(self, metadata_path: Path) -> Optional[List[dict]]:
        """Validate content of metadata file and return mapped content"""
        LOG.info("Fetch SPRING metadata from %s", metadata_path)
        with open(metadata_path, "r") as infile:
            try:
                content = json.load(infile)
            except JSONDecodeError:
                LOG.warning("No content in SPRING metadata file")
                return None
            assert isinstance(content, list)
            metadata = self.mapped_spring_metadata(content)

        if not metadata:
            LOG.warning("Could not find any content in file %s", metadata_path)
            return None

        if len(metadata) != 3:
            LOG.warning("Wrong number of files in SPRING metada file: %s", metadata_path)
            LOG.info("Found %s files, should always be 3 files", len(metadata))
            return None

        return metadata

    def update_metadata_date(self, spring_metadata_path: Path) -> None:
        """Update date in the SPRING metadata file to todays date"""

        today_str = get_date_str(None)
        spring_metadata = self.get_spring_metadata(spring_metadata_path)
        if not spring_metadata:
            raise SyntaxError
        LOG.info("Adding todays date to SPRING metadata file")
        for file_info in spring_metadata:
            file_info["updated"] = today_str

        with open(spring_metadata_path, "w") as outfile:
            outfile.write(json.dumps(spring_metadata))

    @staticmethod
    def mapped_spring_metadata(metadata=List[dict]) -> Optional[List[dict]]:
        """Validate the content of a SPRING metadata file and map them on the schema"""
        file_schema = CrunchyFileSchema(many=True)
        LOG.info("Validating SPRING metadata content")
        try:
            dumped_data = file_schema.load(metadata)
        except ValidationError as err:
            LOG.warning(err.messages)
            return None
        LOG.debug("Spring metadata content looks fine")
        return dumped_data

    @staticmethod
    def get_spring_archive_files(spring_metadata: List[dict]) -> dict:
        """Map the files in SPRING metadata to a dictionary

        Returns: {
                    "fastq_first" : {file_info},
                    "fastq_second" : {file_info},
                    "spring" : {file_info},
                  }
        """
        names_map = {"first_read": "fastq_first", "second_read": "fastq_second", "spring": "spring"}
        archive_files = {}
        for file_info in spring_metadata:
            file_name = names_map[file_info["file"]]
            archive_files[file_name] = file_info
        return archive_files

    @staticmethod
    def get_file_updated_at(spring_metadata: List[dict]) -> datetime.datetime:
        """Check if a SPRING metadata file has been updated and return the date when updated"""
        if "updated" not in spring_metadata[0]:
            return None
        return spring_metadata[0]["updated"]

    @staticmethod
    def check_if_update_spring(file_date: datetime.datetime) -> bool:
        """Check if date is older than FASTQ_DELTA (21 days)"""
        delta = file_date + datetime.timedelta(days=FASTQ_DELTA)
        now = datetime.datetime.now()
        if delta > now:
            LOG.info("FASTQ files are not old enough")
            return False
        return True

    # Methods to get file information
    @staticmethod
    def get_log_dir(file_path: Path) -> Path:
        """Return the path to where logs should be stored"""
        return file_path.parent

    @staticmethod
    def get_sbatch_path(log_dir: Path, compression: str, run_name: str = None) -> Path:
        """Return the path to where sbatch should be printed"""
        if compression == "fastq":
            return log_dir / "_".join([run_name, "compress_fastq.sh"])
        # Only other option is "SPRING"
        return log_dir / "_".join([run_name, "decompress_spring.sh"])

    def _submit_sbatch(self, sbatch_content: str, sbatch_path: Path, pending_path: Path):
        """Submit SLURM job"""
        LOG.info("Submit sbatch")
        if self.dry_run:
            LOG.info("Would submit sbatch %s to slurm", sbatch_path)
            return
        LOG.info("Creating pending flag")
        try:
            pending_path.touch(exist_ok=False)
        except FileExistsError:
            LOG.warning("Pending path exists! Do not submit batch job")
            return

        LOG.debug("Pending flag created")
        with open(sbatch_path, mode="w+t") as sbatch_file:
            sbatch_file.write(sbatch_content)

        sbatch_parameters = [str(sbatch_path.resolve())]
        self.process.run_command(parameters=sbatch_parameters)
        if self.process.stderr:
            LOG.info(self.process.stderr)
        if self.process.stdout:
            LOG.info(self.process.stdout)

    def _get_slurm_header(self, job_name: str, log_dir: str) -> str:
        """Create and return a header for a sbatch script"""
        LOG.info("Generating sbatch header")
        ntasks = 12
        mem = 50
        time = 24
        sbatch_header = SBATCH_HEADER_TEMPLATE.format(
            job_name=job_name,
            account=self.slurm_account,
            log_dir=log_dir,
            conda_env=self.crunchy_env,
            mail_user=self.mail_user,
            ntasks=ntasks,
            time=time,
            mem=mem,
        )
        return sbatch_header

    @staticmethod
    def _get_tmp_dir(prefix: str, suffix: str, base: str = None) -> str:
        """Create a temporary directory and return the path to it"""

        with tempfile.TemporaryDirectory(prefix=prefix, suffix=suffix, dir=base) as dir_name:
            tmp_dir_path = dir_name

        LOG.info("Created temporary dir %s", tmp_dir_path)
        return tmp_dir_path

    @staticmethod
    def _get_slurm_fastq_to_spring(compression_obj: CompressionData) -> str:
        """Create and return the body of a sbatch script that runs FASTQ to SPRING"""
        LOG.info("Generating FASTQ to SPRING sbatch body")
        tmp_dir_path = CrunchyAPI._get_tmp_dir(
            prefix="spring_", suffix="_compress", base=compression_obj.analysis_dir
        )
        sbatch_body = SBATCH_FASTQ_TO_SPRING.format(
            fastq_first=compression_obj.fastq_first,
            fastq_second=compression_obj.fastq_second,
            spring_path=compression_obj.spring_path,
            flag_path=compression_obj.spring_metadata_path,
            pending_path=compression_obj.pending_path,
            tmp_dir=tmp_dir_path,
        )

        return sbatch_body

    @staticmethod
    def _get_slurm_spring_to_fastq(
        compression_obj: CompressionData, checksum_first: str, checksum_second: str
    ) -> str:
        """Create and return the body of a sbatch script that runs SPRING to FASTQ"""
        LOG.info("Generating SPRING to FASTQ sbatch body")
        tmp_dir_path = CrunchyAPI._get_tmp_dir(
            prefix="spring_", suffix="_decompress", base=compression_obj.analysis_dir
        )
        sbatch_body = SBATCH_SPRING_TO_FASTQ.format(
            spring_path=compression_obj.spring_path,
            fastq_first=compression_obj.fastq_first,
            fastq_second=compression_obj.fastq_second,
            pending_path=compression_obj.pending_path,
            checksum_first=checksum_first,
            checksum_second=checksum_second,
            tmp_dir=tmp_dir_path,
        )

        return sbatch_body



"""
    conftest for genotype API
"""

import pytest
from cg.apps.gt import GenotypeAPI
from cg.apps.vogue import VogueAPI


GTCONFIG = {
    "genotype": {
        "database": "database",
        "config_path": "/path/to/config/genotype.yaml",
        "binary_path": "/path/to/genotype_path",
    }
}
VOGUECONFIG = {"vogue": {"binary_path": "/path/to/vogue_path", "config_path": "vogue_config"}}

GENOTYPE_RETURN_SAMPLE = b'{"ACC5346A3": {"status":"pass"}, "SIB903A19": {"status":"pass"}}'

GENOTYPE_RETURN_SAMPLE_ANALYSIS = b'{"ACC5346A3": {"snp": {}}, "SIB903A19": {"snp": {}}}'


@pytest.fixture(scope="function")
def genotype_return():
    """
    genotype config fixture
    """

    _configs = {
        "sample": GENOTYPE_RETURN_SAMPLE,
        "sample_analysis": GENOTYPE_RETURN_SAMPLE_ANALYSIS,
    }

    return _configs


APPTAGS = [
    {"tag": "RMLP10R825", "prep_category": "wgs"},
    {"tag": "METLIFR010", "prep_category": "wes"},
]


class MockApplication:
    """Mock Application"""

    def __init__(self, tag, prep_category):
        self.tag = tag
        self.prep_category = prep_category


class MockApplications:
    """
    Has function to return list of applications
    """

    def __init__(self):
        self.apptag_list = APPTAGS
        self.apptags = []

    def all(self):
        """Returning list of MockApplication instances"""
        for apptag_dict in self.apptag_list:
            apptag = MockApplication(apptag_dict["tag"], apptag_dict["prep_category"])
            self.apptags.append(apptag)
        return self.apptags


class MockStore:
    """
    Mock for Store
    """

    def __init__(self):
        self.apptags = MockApplications()

    def applications(self):
        """Returning MockApplications instance"""
        return self.apptags


@pytest.fixture(scope="function")
def genotype_api():
    """
    genotype API fixture
    """

    _genotype_api = GenotypeAPI(GTCONFIG)
    return _genotype_api


@pytest.fixture(scope="function")
def vogue_api():
    """
    vogue API fixture
    """

    _vogue_api = VogueAPI(VOGUECONFIG)
    return _vogue_api


@pytest.fixture(scope="function")
def store():
    """
    returning MockStore instance
    """

    _store = MockStore()
    return _store



"""Tests for the Scout serialisation models"""

import json

from cg.apps.scout.scout_export import ScoutExportCase


def test_validate_case_father_none(none_case_output: str):
    """Test to validate a case when there are mandatory fields with the value None"""
    cases = json.loads(none_case_output)
    case = cases[0]
    # GIVEN a case that has parent set to None
    assert case["individuals"][0]["father"] is None

    # WHEN validating the output with model
    case_obj = ScoutExportCase(**case)

    # THEN assert father is set to None
    assert case_obj.dict()["individuals"][0]["father"] == "0"


def test_validate_case_parents_none(none_case_output: str):
    """Test to validate a case when there are mandatory fields with the value None"""
    cases = json.loads(none_case_output)
    case = cases[0]
    # GIVEN a case that has parent set to None
    assert case["individuals"][0]["father"] is None
    assert case["individuals"][0]["mother"] is None

    # WHEN validating the output with model
    case_obj = ScoutExportCase(**case)

    # THEN assert father is set to None
    assert case_obj.dict()["individuals"][0]["father"] == "0"
    assert case_obj.dict()["individuals"][0]["mother"] == "0"


def test_validate_empty_diagnosis_phenotypes(none_case_output: str):
    """Test to validate a case when the diagnosis phenotypes is a empty list"""
    cases = json.loads(none_case_output)
    case = cases[0]
    # GIVEN a case that has parent set to None
    assert case["diagnosis_phenotypes"] == []

    # WHEN validating the output with model
    case_obj = ScoutExportCase(**case)

    # THEN assert diagnosis phenotypes is set to None
    assert case_obj.dict()["diagnosis_phenotypes"] == []



"""CLI for delivering files with CG"""

import logging

from typing import List
from pathlib import Path

import click

from cg.meta.deliver import DeliverAPI
from cg.store import Store
from cg.store.models import Family
from cg.apps.hk import HousekeeperAPI
from cg.constants.delivery import (
    PIPELINE_ANALYSIS_OPTIONS,
    PIPELINE_ANALYSIS_TAG_MAP,
)

LOG = logging.getLogger(__name__)


@click.group()
@click.pass_context
def deliver(context):
    """Deliver files with CG."""
    LOG.info("Running CG deliver")
    context.obj["status_db"] = Store(context.obj["database"])
    context.obj["housekeeper_api"] = HousekeeperAPI(context.obj)


@click.command(name="analysis")
@click.option("-c", "--case-id")
@click.option("-t", "--ticket-id", type=int)
@click.option("-d", "--delivery-type", type=click.Choice(PIPELINE_ANALYSIS_OPTIONS), required=True)
@click.option(
    "-i",
    "--inbox",
    help="customer inbox",
)
@click.option("--dry-run", is_flag=True)
@click.pass_context
def deliver_analysis(
    context, case_id: str, ticket_id: int, delivery_type: str, inbox: str, dry_run: bool
):
    """Deliver analysis files to customer inbox"""
    if not (case_id or ticket_id):
        LOG.info("Please provide a case-id or ticket-id")
        return
    inbox = inbox or context.obj.get("delivery_path")
    if not inbox:
        LOG.info("Please specify the root path for where files should be delivered")
        return

    status_db: Store = context.obj["status_db"]
    deliver_api = DeliverAPI(
        store=status_db,
        hk_api=context.obj["housekeeper_api"],
        case_tags=PIPELINE_ANALYSIS_TAG_MAP[delivery_type]["case_tags"],
        sample_tags=PIPELINE_ANALYSIS_TAG_MAP[delivery_type]["sample_tags"],
        project_base_path=Path(inbox),
    )
    deliver_api.set_dry_run(dry_run)
    if case_id:
        case_obj = status_db.family(case_id)
        if not case_obj:
            LOG.warning("Could not find case %s", case_id)
            return
        cases = [case_obj]
    else:
        cases: List[Family] = status_db.get_cases_from_ticket(ticket_id=ticket_id).all()
        if not cases:
            LOG.warning("Could not find cases for ticket_id %s", ticket_id)
            return

    for case_obj in cases:
        deliver_api.deliver_files(case_obj=case_obj)


deliver.add_command(deliver_analysis)



"""The get_links command"""
import logging
from typing import List

import click

from cg.store import Store, models

LOG = logging.getLogger(__name__)


def get_links(
    store: Store, case_id: str = None, sample_id: str = None, ticket: int = None
) -> List[models.FamilySample]:
    """Get link objects for a SAMPLE_ID
    Args:
        case_id(str): petname
        sample_id(str): ACC6395A2
        ticket(int): 123456
    Returns:
        link_objs(list): [models.FamilySample]
    """

    if not any([case_id, sample_id, ticket]):
        LOG.error("Please provide case, sample and/or ticket")
        raise click.Abort

    link_objs = store.links(case_id, sample_id, ticket).all()

    if not link_objs:
        LOG.error("Could not find any links")
        raise click.Abort

    return link_objs



import logging

import click
from cg.store import Store

LOG = logging.getLogger(__name__)


@click.group("reset")
@click.pass_context
def reset_cmd(context):
    """Reset information in the database."""
    context.obj["status_db"] = Store(context.obj["database"])


@reset_cmd.command()
@click.option("-c", "--case_id", help="internal case id, leave empty to process all")
@click.pass_context
def observations(context, case_id):
    """Reset observation links from an analysis to LoqusDB."""

    if case_id:
        observations_uploaded = [context.obj["status_db"].family(case_id)]
    else:
        observations_uploaded = context.obj["status_db"].observations_uploaded()

    for family_obj in observations_uploaded:
        LOG.info("This would reset observation links for: %s", family_obj.internal_id)

    click.confirm("Do you want to continue?", abort=True)

    for family_obj in observations_uploaded:
        context.obj["status_db"].reset_observations(family_obj.internal_id)
        LOG.info("Reset loqus observations for: %s", family_obj.internal_id)

    context.obj["status_db"].commit()



import datetime as dt
import gzip
import logging
import re
from pathlib import Path
from typing import Any, List

from ruamel.yaml import safe_load

from cg.apps.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.mip import parse_trending
from cg.apps.mip.base import MipAPI
from cg.apps.mip.confighandler import ConfigHandler
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI
from cg.apps.tb.models import TrailblazerAnalysis
from cg.constants import (
    COLLABORATORS,
    COMBOS,
    DEFAULT_CAPTURE_KIT,
    FAMILY_ACTIONS,
    MASTER_LIST,
    Pipeline,
)
from cg.exc import CgDataError, LimsDataError
from cg.store import Store, models

LOG = logging.getLogger(__name__)


class MipAnalysisAPI(ConfigHandler, MipAPI):
    """The workflow is accessed through Trailblazer but cg provides additional conventions and
    hooks into the status database that makes managing analyses simpler"""

    def __init__(
        self,
        db: Store,
        hk_api: HousekeeperAPI,
        scout_api: ScoutAPI,
        tb_api: TrailblazerAPI,
        lims_api: LimsAPI,
        script: str,
        pipeline: str,
        conda_env: str,
        root: str,
    ):
        self.db = db
        self.tb = tb_api
        self.hk = hk_api
        self.scout = scout_api
        self.lims = lims_api
        self.script = script
        self.pipeline = pipeline
        self.conda_env = conda_env
        self.root = root

    def check(self, family_obj: models.Family) -> bool:
        """Check stuff before starting the analysis."""
        flowcells = self.db.flowcells(family=family_obj)
        statuses = []
        for flowcell_obj in flowcells:
            LOG.debug("%s: checking flowcell", flowcell_obj.name)
            statuses.append(flowcell_obj.status)
            if flowcell_obj.status == "removed":
                LOG.info("%s: requesting removed flowcell", flowcell_obj.name)
                flowcell_obj.status = "requested"
            elif flowcell_obj.status != "ondisk":
                LOG.warning("%s: %s", flowcell_obj.name, flowcell_obj.status)
        return all(status == "ondisk" for status in statuses)

    @staticmethod
    def get_priority(family_obj: models.Family) -> str:
        """Fetch priority for case id"""
        if not family_obj.priority:
            return "low"
        if isinstance(family_obj.priority, int):
            if family_obj.priority == 0:
                return "low"
            if family_obj.priority > 1:
                return "high"
            return "normal"
        return family_obj.priority

    def get_pedigree_config_path(self, case_id: str) -> Path:
        return Path(self.root, case_id, "pedigree.yaml")

    def get_case_config_path(self, case_id: str) -> Path:
        return Path(self.root, case_id, "analysis", f"{case_id}_config.yaml")

    def pedigree_config(
        self, family_obj: models.Family, pipeline: Pipeline, panel_bed: str = None
    ) -> dict:
        """Make the MIP pedigree config. Meta data for the family is taken from the family object
        and converted to MIP format via trailblazer.

        Args:
            family_obj (models.Family):
            pipeline (str): the name of the pipeline to validate the config against
            panel_bed(str): Overrides default capture kit or bypassed getting panel BED from LIMS

        Returns:
            dict: config_data (MIP format)
        """
        data = self.build_config(family_obj, pipeline=pipeline, panel_bed=panel_bed)

        # Validate and reformat to MIP pedigree config format
        config_data = self.make_pedigree_config(data=data, pipeline=pipeline)
        return config_data

    def get_target_bed_from_lims(self, sample_id: str) -> str:
        """Get target bed filename from lims"""
        target_bed_shortname = self.lims.capture_kit(sample_id)
        if not target_bed_shortname:
            raise LimsDataError("Target bed %s not found in LIMS" % target_bed_shortname)
        bed_version_obj = self.db.bed_version(target_bed_shortname)
        if not bed_version_obj:
            raise CgDataError("Bed-version %s does not exist" % target_bed_shortname)
        return bed_version_obj.filename

    def build_config(
        self, family_obj: models.Family, pipeline: Pipeline, panel_bed: str = None
    ) -> dict:
        """Fetch data for creating a MIP pedigree config file"""

        def get_sample_data(link_obj):
            return {
                "sample_id": link_obj.sample.internal_id,
                "sample_display_name": link_obj.sample.name,
                "analysis_type": link_obj.sample.application_version.application.analysis_type,
                "sex": link_obj.sample.sex,
                "phenotype": link_obj.status,
                "expected_coverage": link_obj.sample.application_version.application.min_sequencing_depth,
            }

        def config_dna_sample(self, link_obj):
            sample_data = get_sample_data(link_obj)
            if sample_data["analysis_type"] == "wgs":
                sample_data["capture_kit"] = panel_bed or DEFAULT_CAPTURE_KIT
            else:
                sample_data["capture_kit"] = panel_bed or self.get_target_bed_from_lims(
                    link_obj.sample.internal_id
                )
            if link_obj.mother:
                sample_data["mother"] = link_obj.mother.internal_id
            if link_obj.father:
                sample_data["father"] = link_obj.father.internal_id
            return sample_data

        def config_rna_sample(self, link_obj):
            sample_data = get_sample_data(link_obj)
            if link_obj.mother:
                sample_data["mother"] = link_obj.mother.internal_id
            if link_obj.father:
                sample_data["father"] = link_obj.father.internal_id
            return sample_data

        dispatch = {
            Pipeline.MIP_DNA: config_dna_sample,
            Pipeline.MIP_RNA: config_rna_sample,
        }

        data = {
            "case": family_obj.internal_id,
            "default_gene_panels": family_obj.panels,
        }
        config_sample = dispatch[pipeline]
        data["samples"] = [config_sample(self, link_obj=link_obj) for link_obj in family_obj.links]
        return data

    @staticmethod
    def fastq_header(line: str) -> dict:
        """handle Illumina's two different header formats
        @see https://en.wikipedia.org/wiki/FASTQ_format

        @HWUSI-EAS100R:6:73:941:1973#0/1

            HWUSI-EAS100R   the unique instrument name
            6   flowcell lane
            73  tile number within the flowcell lane
            941     'x'-coordinate of the cluster within the tile
            1973    'y'-coordinate of the cluster within the tile
            #0  index number for a multiplexed sample (0 for no indexing)
            /1  the member of a pair, /1 or /2 (paired-end or mate-pair reads only)

        Versions of the Illumina pipeline since 1.4 appear to use #NNNNNN
        instead of #0 for the multiplex ID, where NNNNNN is the sequence of the
        multiplex tag.

        With Casava 1.8 the format of the '@' line has changed:

        @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG

            EAS139  the unique instrument name
            136     the run id
            FC706VJ     the flowcell id
            2   flowcell lane
            2104    tile number within the flowcell lane
            15343   'x'-coordinate of the cluster within the tile
            197393  'y'-coordinate of the cluster within the tile
            1   the member of a pair, 1 or 2 (paired-end or mate-pair reads only)
            Y   Y if the read is filtered, N otherwise
            18  0 when none of the control bits are on, otherwise it is an even number
            ATCACG  index sequence


        TODO: add unit test
        """

        rs = {"lane": None, "flowcell": None, "readnumber": None}

        parts = line.split(":")
        if len(parts) == 5:  # @HWUSI-EAS100R:6:73:941:1973#0/1
            rs["lane"] = parts[1]
            rs["flowcell"] = "XXXXXX"
            rs["readnumber"] = parts[-1].split("/")[-1]
        if len(parts) == 10:  # @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG
            rs["lane"] = parts[3]
            rs["flowcell"] = parts[2]
            rs["readnumber"] = parts[6].split(" ")[-1]
        if len(parts) == 7:  # @ST-E00201:173:HCLCGALXX:1:2106:22516:34834/1
            rs["lane"] = parts[3]
            rs["flowcell"] = parts[2]
            rs["readnumber"] = parts[-1].split("/")[-1]

        return rs

    @staticmethod
    def name_file(
        lane: int,
        flowcell: str,
        sample: str,
        read: int,
        undetermined: bool = False,
        date: dt.datetime = None,
        index: str = None,
    ) -> str:
        """Name a FASTQ file following MIP conventions."""
        flowcell = f"{flowcell}-undetermined" if undetermined else flowcell
        date_str = date.strftime("%y%m%d") if date else "171015"
        index = index if index else "XXXXXX"
        return f"{lane}_{date_str}_{flowcell}_{sample}_{index}_{read}.fastq.gz"

    def link_file(self, family: str, sample: str, analysis_type: str, files: List[dict]):
        """Link FASTQ files for a sample."""
        fastq_dir = Path(self.root) / family / analysis_type / sample / "fastq"
        fastq_dir.mkdir(parents=True, exist_ok=True)
        for fastq_data in files:
            fastq_path = Path(fastq_data["path"])
            fastq_name = self.name_file(
                lane=fastq_data["lane"],
                flowcell=fastq_data["flowcell"],
                sample=sample,
                read=fastq_data["read"],
                undetermined=fastq_data["undetermined"],
            )
            dest_path = fastq_dir / fastq_name
            if not dest_path.exists():
                LOG.info(f"linking: {fastq_path} -> {dest_path}")
                dest_path.symlink_to(fastq_path)
            else:
                LOG.debug(f"destination path already exists: {dest_path}")

    def link_sample(self, sample: models.Sample, case_id: str):
        """Link FASTQ files for a sample."""
        file_objs = self.hk.files(bundle=sample.internal_id, tags=["fastq"])
        files = []

        for file_obj in file_objs:
            # figure out flowcell name from header
            with gzip.open(file_obj.full_path) as handle:
                header_line = handle.readline().decode()
                header_info = self.fastq_header(header_line)

            data = {
                "path": file_obj.full_path,
                "lane": int(header_info["lane"]),
                "flowcell": header_info["flowcell"],
                "read": int(header_info["readnumber"]),
                "undetermined": ("_Undetermined_" in file_obj.path),
            }
            # look for tile identifier (HiSeq X runs)
            matches = re.findall(r"-l[1-9]t([1-9]{2})_", file_obj.path)
            if len(matches) > 0:
                data["flowcell"] = f"{data['flowcell']}-{matches[0]}"
            files.append(data)

        self.link_file(
            family=case_id,
            sample=sample.internal_id,
            analysis_type=sample.application_version.application.analysis_type,
            files=files,
        )

    def panel(self, family_obj: models.Family) -> List[str]:
        """Create the aggregated panel file."""
        all_panels = self.convert_panels(family_obj.customer.internal_id, family_obj.panels)
        bed_lines = self.scout.export_panels(all_panels)
        return bed_lines

    def write_panel(self, case_id: str, content: List[str]):
        """Write the gene panel to case dir"""
        out_dir = Path(self.root, case_id)
        out_dir.mkdir(parents=True, exist_ok=True)
        out_path = Path(out_dir, "gene_panels.bed")
        with out_path.open("w") as out_handle:
            out_handle.write("\n".join(content))

    @staticmethod
    def convert_panels(customer: str, default_panels: List[str]) -> List[str]:
        """Convert between default panels and all panels included in gene list."""
        # check if all default panels are part of master list
        if customer in COLLABORATORS and all(panel in MASTER_LIST for panel in default_panels):
            return MASTER_LIST

        # the rest are handled the same way
        all_panels = set(default_panels)

        # fill in extra panels if selection is part of a combo
        for panel in default_panels:
            if panel in COMBOS:
                for extra_panel in COMBOS[panel]:
                    all_panels.add(extra_panel)

        # add OMIM to every panel choice
        all_panels.add("OMIM-AUTO")

        return list(all_panels)

    def _get_latest_raw_file(self, family_id: str, tag: str) -> Any:
        """Get a python object file for a tag and a family ."""

        last_version = self.hk.last_version(bundle=family_id)

        analysis_files = self.hk.files(bundle=family_id, version=last_version.id, tags=[tag]).all()

        if analysis_files:
            analysis_file_raw = self._open_bundle_file(analysis_files[0].path)
        else:
            raise LOG.warning(
                "No post analysis files received from DeliverAPI for '%s'",
                family_id,
            )

        return analysis_file_raw

    def _open_bundle_file(self, relative_file_path: str) -> Any:
        """Open a bundle file and return it as an Python object."""

        full_file_path = Path(self.hk.get_root_dir()).joinpath(relative_file_path)
        open_file = safe_load(open(full_file_path))
        return open_file

    def get_latest_metadata(self, family_id: str) -> dict:
        """Get the latest trending data for a family."""

        mip_config_raw = self._get_latest_raw_file(family_id=family_id, tag="mip-config")
        qcmetrics_raw = self._get_latest_raw_file(family_id=family_id, tag="qcmetrics")
        sampleinfo_raw = self._get_latest_raw_file(family_id=family_id, tag="sampleinfo")
        trending = dict()
        if mip_config_raw and qcmetrics_raw and sampleinfo_raw:
            try:
                trending = parse_trending.parse_mip_analysis(
                    mip_config_raw=mip_config_raw,
                    qcmetrics_raw=qcmetrics_raw,
                    sampleinfo_raw=sampleinfo_raw,
                )
            except KeyError as error:
                LOG.warning(
                    "get_latest_metadata failed for '%s', missing key: %s",
                    family_id,
                    error.args[0],
                )
                LOG.warning(
                    "get_latest_metadata failed for '%s', missing key: %s",
                    family_id,
                    error.args[0],
                )
                trending = dict()
        return trending

    @staticmethod
    def is_dna_only_case(case_obj: models.Family) -> bool:
        """Returns True if all samples of a case has dna application type"""

        for _link in case_obj.links:
            if _link.sample.application_version.application.analysis_type in "wts":
                return False
        return True

    @staticmethod
    def get_skip_evaluation_flag(case_obj: models.Family) -> bool:
        """If any sample in this case is downsampled or external, returns true"""
        for link_obj in case_obj.links:
            downsampled = isinstance(link_obj.sample.downsampled_to, int)
            external = link_obj.sample.application_version.application.is_external
            if downsampled or external:
                LOG.info(
                    "%s: downsampled/external - skip evaluation",
                    link_obj.sample.internal_id,
                )
                return True
        return False

    def set_statusdb_action(self, case_id: str, action: str) -> None:
        if action in [None, *FAMILY_ACTIONS]:
            case_object = self.db.family(case_id)
            case_object.action = action
            self.db.commit()
            return
        LOG.warning(
            f"Action '{action}' not permitted by StatusDB and will not be set for case {case_id}"
        )

    def get_application_type(self, case_id: str) -> str:
        pedigree_config_dict = safe_load(open(self.get_pedigree_config_path(case_id=case_id)))
        analysis_types = set(
            [
                sample.get("analysis_type")
                for sample in pedigree_config_dict["samples"]
                if sample.get("analysis_type")
            ]
        )
        if len(analysis_types) == 1:
            return analysis_types.pop()
        return "wgs"

    def get_case_output_path(self, case_id: str) -> Path:
        return Path(self.root, case_id)

    def get_case_object(self, case_id: str) -> models.Family:
        return self.db.family(case_id)

    def get_analyses_to_clean(self, before: dt.datetime) -> list:
        analyses_to_clean = self.db.analyses_to_clean(pipeline="mip", before=before)
        return analyses_to_clean.all()

    def get_slurm_job_ids_path(self, case_id: str) -> Path:
        return Path(self.get_case_output_path(case_id=case_id), "analysis", "slurm_job_ids.yaml")

    # TrailblazerAPI inherited methods
    def is_latest_analysis_ongoing(self, case_id: str) -> bool:
        return self.tb.is_latest_analysis_ongoing(case_id=case_id)

    def has_latest_analysis_started(self, case_id: str) -> bool:
        return self.tb.has_latest_analysis_started(case_id=case_id)

    def mark_analyses_deleted(self, case_id: str) -> list:
        return self.tb.mark_analyses_deleted(case_id=case_id)

    def add_pending_analysis(
        self,
        case_id: str,
        email: str,
        type: str,
        out_dir: str,
        config_path: str,
        priority: str,
        data_analysis: Pipeline,
    ) -> TrailblazerAnalysis:

        return self.tb.add_pending_analysis(
            case_id=case_id,
            email=email,
            type=type,
            out_dir=out_dir,
            config_path=config_path,
            priority=priority,
            data_analysis=data_analysis,
        )

    def get_latest_analysis_status(self, case_id: str) -> str:
        return self.tb.get_latest_analysis_status(case_id=case_id)

    def get_analyses_from_trailblazer(
        self,
        case_id: str = None,
        query: str = None,
        status: str = None,
        deleted: bool = None,
        temp: bool = False,
        before: dt.datetime = None,
        is_visible: bool = None,
        family: str = None,
        data_analysis: Pipeline = None,
    ) -> list:
        return self.tb.analyses(
            case_id=case_id,
            query=query,
            status=status,
            deleted=deleted,
            temp=temp,
            before=before,
            is_visible=is_visible,
            family=family,
            data_analysis=data_analysis,
        )
